#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_quso_simulation.py.

This file contains the ``QUSOSimulation`` object, which deals with using a
Metropolis algorithm to simulate QUSOs. The ``QUSOSimulation`` object deals
with interfacing with the C code for simulating QUSOs.

"""

from ._simulate_quso import c_simulate_quso as simulate_quso
from itertools import chain
from qubovert import QUSO
from qubovert.utils import QUSOMatrix


__all__ = 'QUSOSimulation',


class QUSOSimulation:
    """QUSOSimulation.

    ``QUSOSimulation`` uses a Metropolis algorithm implemnted in C to simulate
    a QUSO. The QUSO can be "updated" at a temperature ``T``. Thus, we can get
    an idea of the time evolution of a QUSO by creating a temperature
    schedule and seeing how the system evolves.

    Examples
    --------
    Consider the example of the ferromagnetic chain.

    >>> import qubovert as qv
    >>>
    >>> length = 50
    >>> spin_system = sum(
    >>>     -qv.spin_var(i) * qv.spin_var(i+1) * qv.spin_var(i+2)
    >>>     for i in range(length-2)
    >>> )
    >>>
    >>> # initial state is all spin down
    >>> initial_state = {i: -1 for i in range(length)}
    >>> sim = qv.sim.QUSOSimulation(spin_system, initial_state)
    >>>
    >>> # define a schedule. here we simulate at temperature 4 for 25 time
    >>> # steps, then temperature 2 for 25 time steps, then temperature 1 for
    >>> # 10 time steps.
    >>> schedule = (4, 25), (2, 25), (1, 10)
    >>> sim.schedule_update(schedule)
    >>>
    >>> print("final state", sim.state)
    >>> print("last 30 states", sim.get_past_states(30))

    See Also
    --------
    ``qv.sim.PUBOSimulation``, ``qv.sim.PUSOSimulation``,
    ``qv.sim.QUBOSimulation``.

    """

    def __init__(self, L, initial_state=None):
        """__init__.

        Parameters
        ----------
        L : dict, ``qubovert.utils.QUSOMatrix``, or ``qubovert.QUSO`` object.
            The QUSO to simulate. This should map tuples of spin variable
            labels to their respective coefficient in the Hamiltonian.
            For more information, see the docstrings for
            ``qubovert.utils.QUSOMatrix`` and ``qubovert.QUSO``.
        initial_state : dict (optional, defaults to None).
            The initial state to start the simulation in. ``initial_state``
            should map spin label names to their initial values, where each
            value is either 1 or -1. If ``initial_state`` is None, then it
            will be initialized to all 1s.

        """
        # must use type since we don't want errors from inheritance
        if type(L) == QUSOMatrix:
            N = L.max_index + 1
            model = L
            self._mapping = dict(enumerate(range(N)))
            self._reverse_mapping = self._mapping
            self._variables = set(self._mapping.keys())
        elif type(L) != QUSO:
            L = QUSO(L)

        if type(L) == QUSO:
            N = L.num_binary_variables
            model = L.to_quso()
            self._mapping = L.mapping
            self._reverse_mapping = L.reverse_mapping
            self._variables = L.variables

        self._initial_state = (
            initial_state.copy() if initial_state is not None else
            {v: 1 for v in self._variables}
        )
        self._state = [1] * N
        self.set_state(self._initial_state)

        # C arguments
        # create model arrays
        h, num_neighbors = [0.] * N, [0] * N
        neighbors, J = [[] for _ in range(N)], [[] for _ in range(N)]

        for k, v in model.items():
            val = float(v)
            if len(k) == 1:
                h[k[0]] = val
            elif len(k) == 2:
                i, j = k
                neighbors[i].append(j)
                neighbors[j].append(i)
                num_neighbors[i] += 1
                num_neighbors[j] += 1
                J[i].append(val)
                J[j].append(val)
            # ignore offset.
        # flatten the arrays.
        J, neighbors = list(chain(*J)), list(chain(*neighbors))

        self._c_args = h, num_neighbors, neighbors, J

    def __str__(self):
        """__str__.

        Return
        ------
        s : str.

        """
        return self.__class__.__name__

    @property
    def state(self):
        """state.

        A copy of the current state of the system.

        Returns
        -------
        state : dict.
            Dictionary that maps spin labels to their values in {1, -1}.

        """
        return {self._reverse_mapping[k]: v for k, v in enumerate(self._state)}

    @property
    def initial_state(self):
        """initial_state.

        A copy of the initial state of the system.

        Returns
        -------
        initial_state : dict.
            Dictionary that maps binary labels to their values.

        """
        return self._initial_state.copy()

    def set_state(self, state):
        """set_state.

        Set the state of the spin system to ``state``.

        Parameters
        ----------
        state : dict or iterable.
            ``state`` maps the spin variable labels to their corresponding
            values. In other words ``state[v]`` is the value of variable ``v``.
            A value must be either 1 or -1.

        """
        for v in self._variables:
            self._state[self._mapping[v]] = state[v]
        if any(v not in {1, -1} for v in self._state):
            raise ValueError("State must contain only 1's and -1's")

    def reset(self):
        """reset.

        Reset the simulation back to its original state.

        """
        self.set_state(self._initial_state)

    def update(self, T, num_updates=1, seed=None):
        """update.

        Update the simulation at temperature ``T``. Updates the internal state.

        Parameters
        ----------
        T : number >= 0.
            Temperature.
        num_updates : int >= 1 (optional, defaults to 1).
            The number of times to update the simulation at the temperature.
        seed : number (optional, defaults to None).
            The number to seed ``random`` with.

        """
        self.schedule_update([(T, num_updates)], seed)

    def schedule_update(self, schedule, seed=None):
        """schedule_update.

        Update the simulation with a schedule. This function wraps the C
        implementation of the simulation.

        Parameters
        ----------
        schedule : iterable of tuples.
            Each element in ``schedule`` is a pair ``(T, n)`` which designates
            a temperature and a number of updates. See `Notes` below.
        seed : number (optional, defaults to None).
            The number to seed ``random`` with.

        Notes
        -----
        The following two code blocks perform exactly the same thing, although
        the second code block will be faster.

        >>> sim = QUSOSimulation(10)
        >>> for T in (3, 2):
        >>>     sim.update(T, 100)
        >>> sim.update(1, 50)

        >>> sim = QUSOSimulation(10)
        >>> schedule = (3, 100), (2, 100), (1, 50)
        >>> sim.schedule_update(schedule)

        """
        # call the C function
        self._state = simulate_quso(
            self._state,
            *self._c_args,
            schedule,
            seed if seed is not None else -1
        )



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_approximate_extrema.py.

This file functions that estimate the min and max of boolean and spin systems.

"""

__all__ = (
    'approximate_pubo_extrema', 'approximate_puso_extrema',
    'approximate_qubo_extrema', 'approximate_quso_extrema'
)


def approximate_pubo_extrema(P):
    """approximate_pubo_extrema.

    Find the approximate minimum and maximum possible values that a PUBO can
    take. This is very approximate! The estimated minimum value that this
    function returns will definitely be less than or equal to the true
    minimum (most likely, it will be much less). Similarly, the estimated
    maximum value that this function returns will definitely be greater than or
    equal to the true maximum (most likely, it will be much more).

    Parameters
    ----------
    P : dict, PUBO, or PUBOMatrix object.
        ``P`` represents a PUBO. Please see ``qubovert.PUBO`` or
        ``qubovert.utils.PUBOMatrix``.

    Return
    ------
    bounds : tuple (min, max).
        The approximate minimum and maximum of the PUBO.

    """
    min_, max_ = 0, 0
    for k, v in P.items():
        if not k:  # offset
            min_ += v
            max_ += v
        elif v < 0:
            min_ += v
        else:
            max_ += v
    return min_, max_


def approximate_puso_extrema(H):
    """approximate_puso_extrema.

    Find the approximate minimum and maximum possible values that a PUSO can
    take. This is very approximate! The estimated minimum value that this
    function returns will definitely be less than or equal to the true
    minimum (most likely, it will be much less). Similarly, the estimated
    maximum value that this function returns will definitely be greater than or
    equal to the true maximum (most likely, it will be much more).

    Parameters
    ----------
    H : dict, PUSO, or PUSOMatrix object.
        ``H`` represents a PUSO. Please see ``qubovert.PUSO`` or
        ``qubovert.utils.PUSOMatrix``.

    Return
    ------
    bounds : tuple (min, max).
        The approximate minimum and maximum of the PUSO.

    """
    min_, max_ = 0, 0
    for k, v in H.items():
        if not k:  # offset
            min_ += v
            max_ += v
        else:
            min_ -= abs(v)
            max_ += abs(v)
    return min_, max_


def approximate_qubo_extrema(Q):
    """approximate_qubo_extrema.

    Find the approximate minimum and maximum possible values that a QUBO can
    take. This is very approximate! The estimated minimum value that this
    function returns will definitely be less than or equal to the true
    minimum (most likely, it will be much less). Similarly, the estimated
    maximum value that this function returns will definitely be greater than or
    equal to the true maximum (most likely, it will be much more).

    Parameters
    ----------
    Q : dict, QUBO, or QUBOMatrix object.
        ``Q`` represents a QUBO. Please see ``qubovert.QUBO`` or
        ``qubovert.utils.QUBOMatrix``.

    Return
    ------
    bounds : tuple (min, max).
        The approximate minimum and maximum of the QUBO.

    """
    return approximate_pubo_extrema(Q)


def approximate_quso_extrema(L):
    """approximate_quso_extrema.

    Find the approximate minimum and maximum possible values that a QUSO can
    take. This is very approximate! The estimated minimum value that this
    function returns will definitely be less than or equal to the true
    minimum (most likely, it will be much less). Similarly, the estimated
    maximum value that this function returns will definitely be greater than or
    equal to the true maximum (most likely, it will be much more).

    Parameters
    ----------
    L : dict, QUSO, or QUSOMatrix object.
        ``L`` represents a QUSO. Please see ``qubovert.QUSO`` or
        ``qubovert.utils.QUSOMatrix``.

    Return
    ------
    bounds : tuple (min, max).
        The approximate minimum and maximum of the QUSO.

    """
    return approximate_puso_extrema(L)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_qubo_simulation.py.

This file contains the ``QUBOSimulation`` object, which deals with using a
Metropolis algorithm to simulate QUBOs.

"""

from qubovert.utils import qubo_to_quso, spin_to_boolean, boolean_to_spin
from . import QUSOSimulation


__all__ = 'QUBOSimulation',


class QUBOSimulation(QUSOSimulation):
    """QUBOSimulation.

    ``QUBOSimulation`` uses a Metropolis algorithm to simulate a QUBO.
    The QUBO can be "updated" at a temperature ``T``. Thus,
    we can get an idea of the time evolution of a QUBO by creating a
    temperature schedule and seeing how the system evolves.

    ``QUBOSimulation`` inherits from ``QUSOSimulation``. In fact,
    ``QUBOSimulation`` just deals internally with converting to and from
    a spin system; all the simulation is done with ``QUSOSimulation``. See
    ``help(qubovert.sim.QUSOSimulation)`` for more details.

    Examples
    --------
    Consider the following example where we minimize an objective function.

    >>> import qubovert as qv
    >>>
    >>> # create the objective function.
    >>> x = [qv.boolean_var(i) for i in range(10)]
    >>> model = sum(x)
    >>> model.add_constraint_le_zero(x[0] + x[2] - 3 * x[5] - 1, lam=3)
    >>>
    >>> # initial state is all variables equal to 1
    >>> initial_state = {i: 11 for i in range(length)}
    >>> sim = qv.sim.QUBOSimulation(model, initial_state)
    >>>
    >>> # define a schedule. here we simulate at temperature 4 for 25 time
    >>> # steps, then temperature 2 for 25 time steps, then temperature 1 for
    >>> # 10 time steps.
    >>> schedule = (4, 25), (2, 25), (1, 10)
    >>> sim.schedule_update(schedule)
    >>>
    >>> print("final state", sim.state)

    See Also
    --------
    ``qv.sim.PUSOSimulation``, ``qv.sim.QUSOSimulation``,
    ``qv.sim.PUBOSimulation``.

    """

    def __init__(self, Q, initial_state=None):
        """__init__.

        Parameters
        ----------
        Q : dict, ``qubovert.utils.QUBOMatrix``, or ``qubovert.QUBO`` object.
            The QUBO to simulate. This should map tuples of boolean variable
            labels to their respective coefficient in the objective function.
            For more information, see the docstrings for
            ``qubovert.utils.QUBOMatrix`` and ``qubovert.QUBO``.
        initial_state : dict (optional, defaults to None).
            The initial state to start the simulation in. ``initial_state``
            should map boolean label names to their initial values, where each
            value is either 0 or 1. If ``initial_state`` is None, then it
            will be initialized to all 0s.

        """
        model = qubo_to_quso(Q)
        if initial_state is None:
            initial_state = {v: 0 for v in model._variables}
        super().__init__(model, initial_state)

    @property
    def state(self):
        """state.

        A copy of the current state of the system.

        Returns
        -------
        state : dict.
            Dictionary that maps boolean labels to their values in {0, 1}.

        """
        return spin_to_boolean(super().state)

    def set_state(self, state):
        """set_state.

        Set the state of the spin system to ``state``.

        Parameters
        ----------
        state : dict or iterable.
            ``state`` maps the spin variable labels to their corresponding
            values. In other words ``state[v]`` is the value of variable ``v``.
            A value must be either 0 or 1.

        """
        # if we call super then we get the wrong errors
        state = {v: state[v] for v in self._variables}
        if any(v not in {0, 1} for v in state.values()):
            raise ValueError("State must contain only 0's and 1's")
        super().set_state(boolean_to_spin(state))



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_number_partitioning.py.

Contains the NumberPartitioning class. See
``help(qubovert.problems.NumberPartitioning)``.

"""

from qubovert.utils import QUSOMatrix
from qubovert.problems import Problem


__all__ = 'NumberPartitioning',


class NumberPartitioning(Problem):
    """NumberPartitioning.

    Class to manage converting the Number Partitioning problem to and from its
    QUBO and QUSO formluations. Based on the paper "Ising formulations of many
    NP problems", hereforth designated as [Lucas].

    The goal of the NumberPartitioning problem is as follows (quoted from
    [Lucas]):

    Given a list of N positive numbers S = [n1, . . . , nN], is there a
    partition of this set of numbers into two disjoint subsets R and S − R,
    such that the sum of the elements in both sets is the same?

    Note that if we can't do this partitioning, then the next goal is to
    find a partition that almost does this, ie a partition that minimizes
    the difference in the sum between the two partitions.

    This class inherits some methods and attributes from the Problem class. For
    more info, see ``help(qubovert.problems.Problem)``.

    Example
    -------
    >>> from qubovert.problems import NumberPartitioning
    >>> from any_module import qubo_solver
    >>> # or you can use my bruteforce solver...
    >>> # from qubovert.utils import solve_qubo_bruteforce as qubo_solver
    >>> S = [1, 2, 3, 4]
    >>> problem = NumberPartitioning(S)
    >>> Q = problem.to_qubo()
    >>> obj, sol = qubo_solver(Q)
    >>> solution = problem.convert_solution(sol)

    >>> print(solution)
    ([1, 4], [2, 3])  # or ([2, 3], [1, 4])

    >>> print(problem.is_solution_valid(solution))
    True  # since sum([1, 4]) == sum([2, 3])

    >>> print(obj == 0)
    True  # since the solution is valid.

    """

    def __init__(self, S):
        """__init__.

        The goal of the NumberPartitioning problem is as follows (quoted from
        [Lucas]):

        Given a list of N positive numbers S = [n1, . . . , nN], is there a
        partition of this set of numbers into two disjoint subsets R and S − R,
        such that the sum of the elements in both sets is the same?

        Note that if we can't do this partitioning, then the next goal is to
        find a partition that almost does this, ie a partition that minimizes
        the difference in the sum between the two partitions.

        All naming conventions follow the names in the paper [Lucas].

        Parameters
        ----------
        S: tuple or list.
            tuple or list of N nonzero numbers that we are attempting to
            partition into two partitions of equal sum.

        Example
        -------
        >>> problem = NumberPartitioning([1, 2, 3, 4])

        """
        if not all(S):
            raise ValueError("Numbers must be nonzero")
        self._input_type = type(S)
        self._S = self._input_type(x for x in S)  # copy the input
        self._N = len(S)

    @property
    def S(self):
        """S.

        A copy of the S list. Updating the copy will not update the instance
        list.

        Return
        ------
        S : tuple or list.
            A copy of the list of numbers defining the partitioning problem.

        """
        return self._input_type(x for x in self._S)

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        The number of binary variables that the QUBO and QUSO use.

        Return
        ------
        num : int.
            The number of variables in the QUBO/QUSO formulation.

        """
        return self._N

    def to_quso(self, A=1):
        r"""to_quso.

        Create and return the number partitioning problem in QUSO form
        following section 2.1 of [Lucas]. It is
        the value such that the solution to the QUSO formulation is 0
        if a valid number partitioning exists.

        Parameters
        ----------
        A: positive float (optional, defaults to 1).
            Factor in front of objective function. See section 2.1 of [Lucas].

        Return
        ------
        L : qubovert.utils.QUSOMatrix object.
            For most practical purposes, you can use QUSOMatrix in the
            same way as an ordinary dictionary. For more information, see
            ``help(qubovert.utils.QUSOMatrix)``.

        Example
        -------
        >>> problem = NumberPartitioning([1, 2, 3, 4])
        >>> L = problem.to_quso()

        """
        L = QUSOMatrix({(i,): self._S[i] for i in range(self._N)})
        return A * L * L

    def convert_solution(self, solution, spin=False):
        """convert_solution.

        Convert the solution to the QUBO or QUSO to the solution to the
        Number Partitioning problem.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        res: tuple of iterables (partition1, partition2).
            partition1 : list, tuple, or iterable.
                The first partition. If the inputted S is a tuple, then
                partition1 will as a tuple, otherwise it will be a list.
            partition2 : list, tuple, or iterable.
                The second partition.

        Example
        -------
        >>> problem = NumberPartitioning([1, 2, 3, 4])
        >>> Q = problem.to_qubo()
        >>> value, solution = solve_qubo(Q)
        >>> print(solution)
        [0, 1, 1, 0]
        # ie 1 and 4 are in one partition, and 2 and 3 are in the other

        >>> print(problem.convert_solution(solution))
        ([1, 4], [2, 3])

        """
        if not isinstance(solution, dict):
            solution = dict(enumerate(solution))
        partition1 = self._input_type(
            self._S[i] for i, v in solution.items() if v == 1
        )
        partition2 = self._input_type(
            self._S[i] for i, v in solution.items() if v != 1
        )
        return partition1, partition2

    def is_solution_valid(self, solution, spin=False):
        """is_solution_valid.

        Returns whether or not the proposed solution partitions S into two sets
        of equal sum.

        Parameters
        ----------
        solution : iterable or dict.
            solution can be the output of NumberPartitioning.convert_solution,
            or the  QUBO or QUSO solver output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        valid : boolean.
            True if the proposed solution is valid, else False.

        """
        not_converted = (
            not isinstance(solution, tuple) or len(solution) != 2 or
            not isinstance(solution[0], self._input_type) or
            not isinstance(solution[1], self._input_type)
        )

        if not_converted:
            solution = self.convert_solution(solution, spin)

        return sum(solution[0]) == sum(solution[1])



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_job_sequencing.py.

Contains the JobSequencing class.
See ``help(qubovert.problems.JobSequencing)``.

"""

from numpy import log2
from qubovert.utils import (
    QUBOMatrix, decimal_to_boolean, solution_type, spin_to_boolean
)
from qubovert.problems import Problem


__all__ = 'JobSequencing',


class JobSequencing(Problem):
    """JobSequencing.

    Class to manage converting Job Sequencing to and from its QUBO and
    QUSO formluations. Based on the paper "Ising formulations of many
    NP problems", hereforth designated as [Lucas].

    The goal of the JobSequencing problem is as follows. Given workers and
    jobs, where each job has a designated length, assign each of the jobs to
    one of the workers such that largest total length assigned to a worker is
    minimized.

    This class inherits some methods and attributes from the
    ``qubovert.problems.Problem`` class.

    Example
    -------
    >>> from qubovert.problems import JobSequencing
    >>> from any_module import qubo_solver
    >>> # or you can use my bruteforce solver...
    >>> # from qubovert.utils import solve_qubo_bruteforce as qubo_solver

    >>> job_lengths = {"job1": 2, "job2": 3, "job3": 1}
    >>> num_workers = 2
    >>> problem = JobSequencing(job_lengths, num_workers)
    >>> Q = problem.to_qubo()
    >>> obj, sol = qubo_solver(Q)
    >>> solution = problem.convert_solution(sol)

    >>> print(solution)
    ({'job1', 'job3'}, {'job2'})  # or ({'job2'}, {'job1', 'job3'})
    >>> print(problem.is_solution_valid(solution))
    True  # since each job is covered exactly once
    >>> print(
            obj ==
            max(sum(job_lengths[i] for i in x) for x in solution) ==
            3
        )
    True

    """

    def __init__(self, job_lengths, num_workers, log_trick=True, M=None):
        """__init__.

        The goal of the JobSequencing problem is as follows. Given workers and
        jobs, where each job has a designated length, assign each of the jobs
        to  one of the workers such that largest total length assigned to a
        worker is minimized.

        Parameters
        ----------
        job_lengths : tuple or list of integers, or a dict.
            The length of each job. If ``job_lengths`` is a ``dict`` then
            each key is a job that maps to its length. Otherwise,
            the index of ``job_lengths`` is the job and it is mapped to its
            length. Note that the lengths must be integers!
        num_workers : int.
            The number of workers that can be assigned jobs.
        log_trick : boolean (optional, defaults to True).
            Indicates whether or not to use the log trick discussed in [Lucas].
        M : int (optional, defaults to None). We recommend not adjusting this.
            The maximum expected numbers of jobs assigned to a single worker,
            see [Lucas]. If M is None, then it will be set to the value
            required to ensure accuracy of the model. To possibly sacrifice
            accuracy but decrease the number of variables in the model, you can
            set M to something smaller.

        """
        self._log_trick = log_trick
        self._m = int(num_workers)

        if isinstance(job_lengths, dict):
            self._lengths = job_lengths.copy()
            self._input_type = dict
        else:
            self._lengths = dict(enumerate(job_lengths))
            self._input_type = type(job_lengths)

        self._max_L = max(self._lengths.values())
        self._N = len(self._lengths)
        self._M = self._N * self._max_L if M is None else M
        self._log_M = int(log2(self._M)) + 1

        # map job names to integers
        self._job_to_int = {v: k for k, v in enumerate(self._lengths.keys())}

    @property
    def job_lengths(self):
        """job_lengths.

        A copy of the job lengths.

        Return
        ------
        job_lengths : tuple or list of integers, or dict.
            Will be whichever type was inputted in the initialization of the
            class. The length of each job. If ``job_lengths`` is a ``dict``
            then each key is a job that maps to its length. Otherwise,
            the index of ``job_lengths`` is the job and it is mapped to its
            length.

        """
        if self._input_type == dict:
            return self._lengths.copy()
        else:
            return self._input_type(
                self._lengths[x] for x in sorted(self._lengths.keys())
            )

    @property
    def num_workers(self):
        """num_workers.

        A copy of the number of workers.

        Return
        ------
        num_workers : int.
            The number of workers allowed for the problem.

        """
        return self._m

    @property
    def log_trick(self):
        """log_trick.

        A copy of the value for log_trick, indicating whether or not to use
        the log trick. See [Lucas].

        Return
        ------
        log_trick : bool.

        """
        return self._log_trick

    @property
    def M(self):
        """M.

        A copy of the ``M`` constant.

        Return
        ------
        M : int.
            If ``M`` was supplied in the initialization of the class, then this
            will return the same. Otherwise, it will return the value that the
            class calculated for its default.

        """
        return self._M

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        The number of binary variables that the QUBO and QUSO use.

        Return
        ------
        num :  int.
            The number of variables in the QUBO/QUSO formulation.

        """
        if self._log_trick:
            return self._m * self._N + (self._m - 1) * self._log_M
        return self._m * self._N + (self._m - 1) * self._M

    def _x(self, job, worker):
        """_x.

        Return the integer index for the job ``job`` and the worker ``worker``.
        This is equivalent to the ``x`` binary variable in [Lucas].

        Parameters
        ----------
        job : key of self._lengths.
            Job to be covered.
        worker : int.
            Worker to cover the job. ``0 <= worker < self._m``

        Return
        ------
        x : int >= 0.
            Index mapping of the binary variable.

        """
        # same regardless of log trick
        return self._job_to_int[job] * self._m + worker

    def _y(self, i, worker):
        """_y.

        Return the integer index for the slack variables ``y`` in [Lucas].
        The ``y``'s are defined to be >= 0.

        Parameters
        ----------
        i : int.
            Index of the slack variable.
        worker : int.
            Worker to cover the job. ``0 <= worker < self._m``

        Return
        ------
        y : int >= max(self._x(job, worker)).
            Index mapping of the slack variable.

        """
        # same regardless of log trick
        # all the minuses are because we don't need a y(i, 0)!
        return self._N * self._m + i * (self._m - 1) + worker - 1

    def to_qubo(self, A=None, B=1):
        r"""to_qubo.

        Create and return the job sequencing problem in QUBO form following
        section 6.3 of [Lucas]. The Q matrix for the QUBO
        will be returned as an uppertriangular dictionary. Thus, the problem
        becomes minimizing :math:`\sum_{i \leq j} x_i x_j Q_{ij}`. A and B are
        parameters to enforce constraints.

        If all the constraints are satisfied, then the objective function will
        be equal to the total length of the scheduling.

        Parameters
        ----------
        A: positive float (optional, defaults to None).
            ``A`` enforces the constraints. If ``A is None``, then ``A`` will
            be chosen such that the minimum of the QUBO is `guaranteed` to
            satisfy the constraints (``A = B*max(L)``). This may not be the
            best value for any particular QUBO solver, since it may cause a non
            smooth landscape that is hard to minimize.
        B: positive float (optional, defaults to 1).
            ``B`` is the constant in front of the portion of the QUBO to
            minimize.

        Return
        ------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, a QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUBOMatrix)``.

        """
        # all naming conventions follow the paper listed in the docstring

        if A is None:
            A = B * self._max_L

        Q = QUBOMatrix()

        Q += self._N * A  # offset comes from the first constraint

        # encode H_B (equation 55)
        # minimize worker 0's length
        for job, length in self._lengths.items():
            ind = self._x(job, 0)  # worker zero
            Q[(ind,)] += B * length

        # encode H_A (equation 54)

        # enforce that each job is covered exactly once.
        for job in self._lengths:
            for worker in range(self._m):
                ind = self._x(job, worker)
                Q[(ind,)] -= 2 * A
                for workerp in range(self._m):
                    indp = self._x(job, workerp)
                    Q[(ind, indp)] += A

        # enforce worker 0's length is larger than all the other workers'
        # lengths
        max_M = self._log_M if self._log_trick else self._M
        for worker in range(1, self._m):  # exclude worker 0

            for n in range(max_M):
                ind = self._y(n, worker)
                for np in range(max_M):
                    indp = self._y(np, worker)
                    if self._log_trick:
                        Q[(ind, indp)] += A * pow(2, n + np)
                    else:
                        Q[(ind, indp)] += A * (n + 1) * (np + 1)

                for job, length in self._lengths.items():
                    ind1, ind2 = self._x(job, worker), self._x(job, 0)
                    val = 2 * A * length * (
                        pow(2, n) if self._log_trick else n + 1
                    )
                    Q[(ind, ind1)] += val
                    Q[(ind, ind2)] -= val

            for job, length in self._lengths.items():
                ind1, ind2 = self._x(job, worker), self._x(job, 0)
                for jobp, lengthp in self._lengths.items():
                    ind1p, ind2p = self._x(jobp, worker), self._x(jobp, 0)
                    val = A * length * lengthp
                    Q[(ind1, ind1p)] += val
                    Q[(ind2, ind2p)] += val
                    Q[(ind2, ind1p)] -= val
                    Q[(ind1, ind2p)] -= val

        return Q

    def convert_solution(self, solution, spin=False):
        """convert_solution.

        Convert the solution to the QUBO or QUSO to the solution to the Job
        Sequencing problem.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Returns
        -------
        res : tuple of sets.
            Each element of the tuple corresponds to a worker. Each element
            of the tuple is a set of jobs that are assigned to that worker.

        """
        sol_type = solution_type(solution, 'spin' if spin else 'bool')
        if sol_type == 'spin':
            solution = spin_to_boolean(solution)
        res = tuple(set() for _ in range(self._m))
        for worker in range(self._m):
            for job in self._lengths:
                if solution[self._x(job, worker)] == 1:
                    res[worker].add(job)
        return res

    def is_solution_valid(self, solution, spin=False):
        """is_solution_valid.

        Returns whether or not the proposed solution completes all the jobs
        exactly once.

        Parameters
        ----------
        solution : iterable or dict.
            solution can be the output of JobSequencing.convert_solution,
            or the  QUBO or QUSO solver output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        valid : boolean.
            True if the proposed solution is valid, else False.

        """
        converted = isinstance(solution, tuple) and all(
            isinstance(x, set) for x in solution
        )
        if not converted:
            solution = self.convert_solution(solution, spin)

        all_jobs, completed_jobs = set(self._lengths.keys()), set()

        for worker in solution:
            for job in worker:
                if job in completed_jobs:  # job covered more than once.
                    return False
                completed_jobs.add(job)

        return completed_jobs == all_jobs  # all jobs must be covered.

    def solve_bruteforce(self, all_solutions=False):
        """solve_bruteforce.

        Solves the JobSequence problem exactly with a brute force method. THIS
        SHOULD NOT BE USED FOR LARGE PROBLEMS! The advantage over this method
        as opposed to using a brute force QUBO solver is that the QUBO
        formulation has many slack variables.

        Parameters
        ----------
        all_solutions : boolean (optional, defaults to False).
            If ``all_solutions`` is set to True, all the best solutions to the
            problem will be returned rather than just one of the best. If the
            problem is very big, then it is best if ``all_solutions == False``,
            otherwise this function will use a lot of memory.

        Returns
        -------
        res : tuple of sets or list of tuple of sets.
            Each element of the tuple corresponds to a worker. Each element
            of the tuple is a set of jobs that are assigned to that worker. If
            ``all_solutions == True`` then ``res`` will be a list, where each
            element of the list will be one of the optimal solutions.

        Example
        -------
        >>> job_lengths = {"job1": 2, "job2": 3, "job3": 1}
        >>> num_workers = 2
        >>> problem = JobSequencing(job_lengths, num_workers)
        >>> print(problem.solve_bruteforce())
        ({'job1', 'job3'}, {'job2'})  # or ({'job2'}, {'job1', 'job3'})

        >>> print(problem.solve_bruteforce(True))
        [({'job1', 'job3'}, {'job2'}), ({'job2'}, {'job1', 'job3'})]

        """
        best = None, None
        all_sols, n = {}, self._m * self._N
        for i in range(1 << n):
            sol = self.convert_solution(decimal_to_boolean(i, n))
            if self.is_solution_valid(sol):
                obj = max(
                    sum(self._lengths[job] for job in cluster)
                    for cluster in sol
                )
                if not all_solutions and (best[0] is None or obj < best[0]):
                    best = obj, sol
                elif all_solutions and (best[0] is None or obj <= best[0]):
                    best = obj, sol
                    all_sols.setdefault(obj, []).append(sol)

        if all_solutions:
            return all_sols[best[0]]
        return best[1]



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_annea_resultsl.py.

This file contains the objects to deal with anneal results.

"""

from qubovert.utils import spin_to_boolean, boolean_to_spin


__all__ = 'AnnealResult', 'AnnealResults'


class AnnealResult:
    """AnnealResult.

    This class deals with an individual result from an anneal. See the below
    example.

    Example
    -------
    In this example, ``res`` represents a state the gives value ``-2``,
    and is part of a spin model.

    >>> from qubovert.sim import AnnealResult
    >>>
    >>> res = AnnealResult({0: 1, 1: -1, 2: -1}, -2, True)
    >>> print(res.state)
    {0: 1, 1: -1, 2: -1}
    >>> print(res.value)
    -2
    >>> print(res.spin)
    True

    We can convert it to a boolean state with

    >>> boolean_res = res.to_boolean()
    >>> print(boolean_res.state)
    {0: 0, 1: 1, 2: 1}
    >>> print(boolean_res.value)
    -2
    >>> print(boolean_res.spin)
    False

    """

    def __init__(self, state, value, spin):
        """__init__.

        Parameters
        ----------
        state : dict.
            Maps binary labels to their values.
        value : number.
            The value of the model with ``state``.
        spin : bool.
            Indicates whether ``state`` came from a boolean or spin model.

        """
        self.state, self.value, self.spin = state, value, spin

    def __eq__(self, other):
        """__eq__.

        Determine if two AnnealResult objects are equivalent.

        Parameters
        ----------
        other : AnnealResult object.

        Returns
        -------
        res : bool.

        """
        return all((
            self.state == other.state,
            self.value == other.value,
            self.spin == other.spin
        ))

    def copy(self):
        """copy.

        Return a copy of ``self``.

        Returns
        -------
        res : AnnealResult.

        """
        return AnnealResult(self.state.copy(), self.value, self.spin)

    def to_boolean(self):
        """to_boolean.

        Convert the result to a boolean result.

        Returns
        -------
        res : AnnealResult object.
            A boolean version of ``self``. If ``self.spin == False``, then
            ``res`` will be the same as ``self``.

        """
        if not self.spin:
            return self.copy()
        return AnnealResult(spin_to_boolean(self.state), self.value, False)

    def to_spin(self):
        """to_spin.

        Convert the result to a spin result.

        Returns
        -------
        res : AnnealResult object.
            A spin version of ``self``. If ``self.spin == True``, then
            ``res`` will be the same as ``self``.

        """
        if self.spin:
            return self.copy()
        return AnnealResult(boolean_to_spin(self.state), self.value, True)

    def __str__(self):
        """__str__.

        Return the string representation of ``self``.

        Return
        ------
        s : str.

        """
        return "  state: %s\n  value: %g\n   spin: %s" % (
            self.state, self.value, self.spin
        )

    def __repr__(self):
        """__repr__.

        Create the representation of ``self``.

        Return
        ------
        r : str.

        """
        return "AnnealResult(state=%s, value=%g, spin=%s)" % (
            self.state, self.value, self.spin
        )


class AnnealResults(list):
    """AnnealResults.

    An object to manage accessing the results of the simulated annealing
    functions. ``AnnealResults`` is a subclass of ``list``.

    Let ``res`` be the output of one of the simulated annealing functions. Then
    a user can access the best result with ``res.best``. The best state will
    be ``res.best.state`` and its corresponding value ``res.best.value``.

    The user can iterate through all of the results with

    >>> for r in res:
    >>>     print(r.state, r.value)

    The user can sort the results from best to worst with
    ``res.sort_by_value()``. Then iterating through ``res`` will be in that
    order.

    The user can also convert each result state to/from boolean and spin
    states.

    """

    def __init__(self, spin):
        """__init__.

        Parameters
        ----------
        spin : bool.
            Indicates whether the results are coming from a boolean or spin
            model.

        """
        super().__init__()
        self._spin, self.best = spin, None

    @property
    def spin(self):
        """spin.

        Return
        ------
        spin : bool.
            Whether ``self`` contains spin or boolean results.

        """
        return self._spin

    def copy(self):
        """copy.

        Return
        ------
        res : AnnealResults object.
            A deep copy of ``self``.

        """
        return AnnealResults.from_list(self, self._spin)

    def add_state(self, state, value):
        """add_state.

        Add the state to the record.

        Parameters
        ----------
        state : dict.
            Maps variable labels to their values.
        value : number.
            The value of the model with ``state``.

        Returns
        -------
        None.

        """
        self.append(AnnealResult(state, value, self._spin))

    def append(self, result):
        """append.

        Add the result to the record.

        Parameters
        ----------
        result : AnnealResult object.
            See ``help(qubovert.sim.AnnealResult)``.

        Returns
        -------
        None.

        """
        if self.best is None or result.value < self.best.value:
            self.best = result
        super().append(result)

    def to_boolean(self):
        """to_boolean.

        Convert each result to a boolean result. Note that if ``self.spin`` is
        False, then this function will just return a copy of ``self``.

        Returns
        -------
        res : AnnealResults object.

        """
        res = AnnealResults(False)
        for r in self:
            res.append(r.to_boolean())
        return res

    def to_spin(self):
        """to_spin.

        Convert each result to a spin result. Note that if ``self.spin`` is
        True, then this function will just return a copy of ``self``.

        Returns
        -------
        res : AnnealResults object.

        """
        res = AnnealResults(True)
        for r in self:
            res.append(r.to_spin())
        return res

    def sort_by_value(self):
        """sort_by_value.

        Sort the results in ``self`` in increasing order of their values.

        """
        self.sort(key=lambda x: x.value)

    def __str__(self):
        """__str__.

        Return the string representation of ``self``.

        Returns
        -------
        s : str.

        """
        return "AnnealResults\n" + "\n\n".join(str(x) for x in self)

    def __getitem__(self, index):
        """__getitem__.

        Override ``list.__getitem__`` so that when slicing we return a
        ``qubovert.sim.AnnealResults`` item instead of a list.

        Parameters
        ----------
        index : int or slice object.

        Returns
        -------
        res : qubovert.sim.AnnealResults object.

        """
        res = super().__getitem__(index)
        if isinstance(index, slice):
            res = AnnealResults.from_list(res, self._spin)
        return res

    def clear(self):
        """clear.

        Override ``list.clear`` so that it also removes ``self.best``.

        """
        self.best = None
        super().clear()

    @staticmethod
    def from_list(l, spin):
        """from_list.

        Create an ``AnnealResults`` object from a list of ``AnnealResult``
        objects.

        Parameters
        ----------
        l : list.

        Returns
        -------
        res : qubovert.sim.AnnealResults object.

        """
        res = AnnealResults(spin)
        for i in l:
            res.append(i.copy())
        return res

    def filter(self, func):
        """filter.

        Return a new AnnealResults object whose elements are filtered by the
        function ``func``. ``func`` takes in a ``qubovert.sim.AnnealResult``
        object and returns a boolean indicating whether it should remain
        in the filtered results.

        Parameters
        ----------
        func : function.
            ``func`` takes in a ``qubovert.sim.AnnealResult`` object and
            returns a boolean indicating whether it should remain in the
            filtered results.

        Returns
        -------
        res : qubovert.sim.AnnealResults object.

        Example
        -------
        >>> import qubovert as qv
        >>>
        >>> model = qv.boolean_var(0) * qv.boolean_var(1)
        >>> anneal_res = qv.sim.anneal_qubo(model, num_anneals=3)
        >>>
        >>> anneal_res
        [AnnealResult(state={0: 0, 1: 1}, value=0, spin=False),
         AnnealResult(state={0: 1, 1: 0}, value=0, spin=False),
         AnnealResult(state={0: 0, 1: 0}, value=0, spin=False)]
        >>> filtered_anneal_res = anneal_res.filter(
        >>>     lambda x: x.state[0] == 0
        >>> )
        >>> filtered_anneal_res
        [AnnealResult(state={0: 0, 1: 1}, value=0, spin=False),
         AnnealResult(state={0: 0, 1: 0}, value=0, spin=False)]

        """
        return AnnealResults.from_list(
            filter(func, self),
            self._spin
        )

    def filter_states(self, func):
        """filter_states.

        Return a new AnnealResults object whose states are filtered by the
        function ``func``. ``func`` takes in a ``dict`` representing a state
        and returns a boolean indicating whether it should remain in the
        filtered results.

        Parameters
        ----------
        func : function.
            ``func`` takes in a ``dict`` representing a state and
            returns a boolean indicating whether it should remain in the
            filtered results.

        Returns
        -------
        res : qubovert.sim.AnnealResults object.

        Example
        -------
        >>> import qubovert as qv
        >>>
        >>> model = qv.boolean_var(0) * qv.boolean_var(1)
        >>> anneal_res = qv.sim.anneal_qubo(model, num_anneals=3)
        >>>
        >>> anneal_res
        [AnnealResult(state={0: 0, 1: 1}, value=0, spin=False),
         AnnealResult(state={0: 1, 1: 0}, value=0, spin=False),
         AnnealResult(state={0: 0, 1: 0}, value=0, spin=False)]
        >>> filtered_anneal_res = anneal_res.filter_states(
        >>>     lambda x: x[0] == 0
        >>> )
        >>> filtered_anneal_res
        [AnnealResult(state={0: 0, 1: 1}, value=0, spin=False),
         AnnealResult(state={0: 0, 1: 0}, value=0, spin=False)]

        """
        return AnnealResults.from_list(
            filter(lambda x: func(x.state), self),
            self._spin
        )

    def apply_function(self, func):
        """apply_function.

        Apply the function ``func`` to each element in ``self`` to create a
        new version of ``self``.

        Parameters
        ----------
        func : function.
            ``func`` takes in a ``qubovert.sim.AnnealResult`` object and
            returns a ``qubovert.sim.AnnealResult`` object.

        Returns
        -------
        res : qubovert.sim.AnnealResults object.

        Example
        -------
        >>> import qubovert as qv
        >>>
        >>> model = qv.boolean_var('a') * qv.boolean_var('b')
        >>> qubo = model.to_qubo()
        >>> anneal_res = qv.sim.anneal_qubo(qubo, num_anneals=3)
        >>> anneal_res
        [AnnealResult(state={0: 0, 1: 1}, value=0, spin=False),
         AnnealResult(state={0: 0, 1: 0}, value=0, spin=False),
         AnnealResult(state={0: 0, 1: 0}, value=0, spin=False)]
        >>> new_res = anneal_res.apply_function(
                lambda x: qv.sim.AnnealResult(
                    model.convert_solution(x.state), x.value, x.spin
                )
            )
        >>> new_res
        [AnnealResult(state={'a': 0, 'b': 1}, value=0, spin=False),
         AnnealResult(state={'a': 0, 'b': 0}, value=0, spin=False),
         AnnealResult(state={'a': 0, 'b': 0}, value=0, spin=False)]

        """
        return AnnealResults.from_list(
            [func(x) for x in self],
            self._spin
        )

    def convert_states(self, func):
        """convert_states.

        Apply the function ``func`` to each state in ``self`` to create a
        new version of ``self``.

        Parameters
        ----------
        func : function.
            ``func`` takes in a dict that maps variable names to their values,
            and returns a dict.

        Returns
        -------
        res : qubovert.sim.AnnealResults object.

        Example
        -------
        >>> import qubovert as qv
        >>>
        >>> model = qv.boolean_var('a') * qv.boolean_var('b')
        >>> qubo = model.to_qubo()
        >>> anneal_res = qv.sim.anneal_qubo(qubo, num_anneals=3)
        >>> anneal_res
        [AnnealResult(state={0: 0, 1: 1}, value=0, spin=False),
         AnnealResult(state={0: 0, 1: 0}, value=0, spin=False),
         AnnealResult(state={0: 0, 1: 0}, value=0, spin=False)]
        >>> new_res = anneal_res.convert_states(model.convert_solution)
        >>> new_res
        [AnnealResult(state={'a': 0, 'b': 1}, value=0, spin=False),
         AnnealResult(state={'a': 0, 'b': 0}, value=0, spin=False),
         AnnealResult(state={'a': 0, 'b': 0}, value=0, spin=False)]

        """
        return self.apply_function(
            lambda x: AnnealResult(func(x.state), x.value, x.spin)
        )

    def __add__(self, other):
        """__add__.

        Override ``list.__add__`` to return a ``AnnealResults`` object.

        Parameters
        ----------
        other : qubovert.sim.AnnealResults object.

        Returns
        -------
        res : qubovert.sim.AnnealResults object.

        """
        return AnnealResults.from_list(super().__add__(other), self._spin)

    def __mul__(self, other):
        """__mul__.

        Override ``list.__mul__`` to return a ``AnnealResults`` object.

        Parameters
        ----------
        other : qubovert.sim.AnnealResults object.

        Returns
        -------
        res : qubovert.sim.AnnealResults object.

        """
        return AnnealResults.from_list(super().__mul__(other), self._spin)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``np`` contains many NP problems.

This module contains many NP problems. We organize all the problems into
categories. But we import them globally for user use. See the ``__all__`` value
for all the problems imported. Most conversions to QUBO/QUSO form are based
on [Lucas]_.

References
----------
.. [Lucas] Andrew Lucas. Ising formulations of many np problems. Frontiers in
   Physics, 2:5, 2014.

"""

from .bilp import *
from .coloring import *
from .covering import *
from .cycles import *
from .packing import *
from .partitioning import *
from .tree import *

from .bilp import __all__ as __all_bilp__
from .coloring import __all__ as __all_coloring__
from .covering import __all__ as __all_covering__
from .cycles import __all__ as __all_cycles__
from .packing import __all__ as __all_packing__
from .partitioning import __all__ as __all_partitioning__
from .tree import __all__ as __all_tree__

__all__ = (
    __all_bilp__ + __all_coloring__ + __all_covering__ + __all_cycles__ +
    __all_packing__ + __all_partitioning__ + __all_tree__
)

name = "np"



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_problem_parentclass.py.

This file contains the class Problem, which is the parent class to all
the problem classes.

"""

from qubovert.utils import Conversions, PUBOMatrix


__all__ = 'Problem',


class Problem(Conversions):
    """Problem.

    This acts a parent class to all the QUBO and QUSO conversion problem
    classes. The ``__new__`` method keeps track of the problem args. The
    ``repr`` method uses those input args, such that
    ``eval(repr(cls)) == cls``. Finally, we define a ``__eq__`` method to
    determine if two problems are the same. The rest of the methods are to be
    implemented in child classes.

    Subclasses `must` define at least one of the following methods:
        ``to_qubo``, ``to_quso``.

    If at least one of them is defined, then they will both work. The same is
    true for ``to_pubo`` and ``to_puso``.
    ``Problem`` inherits from ``Conversions``, for more details see
    ``help(qubovert.utils.Conversions)``
    """

    def __new__(cls, *args, **kwargs):
        """__new__.

        Creates the object and keeps track of the input arguments and keyword
        arguments. Child classes should not change this. This method will be
        called before every __init__ is called. We use __new__ to keep track
        of input arguments instead of using __init__ so that child class
        implementations don't have to worry about it. Ie child classes
        don't have to call `super().__init__(*args, **kwargs)` in their
        __init__ method.

        Parameters
        ----------
        arguments : Defined in child classes.

        Return
        ------
        obj : instance of the child class.

        """
        obj = super().__new__(cls)
        obj._problem_args, obj._problem_kwargs = args, kwargs.copy()
        return obj

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        The number of binary variables that the QUBO/QUSO uses. Should be
        implemented in the child class.

        Return
        ------
        num : int.
            The number of variables in the QUBO/QUSO formulation.

        """
        raise NotImplementedError("To be implemented in the child class")

    def __repr__(self):
        """__repr__.

        Same as __str__, but will give a truncted output.

        Return
        ------
        s : str.

        """
        s = str(self)
        if len(s) < 50:
            return s
        return s[:50] + " ...)"

    def __str__(self):
        """__str__.

        Defined such that the following is true (assuming you have imported
        * from qubovert.problems).

        >>> s = Class_derived_from_Problem(*args)
        >>> eval(str(s)) == s
        True

        Return
        ------
        s : str.

        """
        s = self.__class__.__name__ + "("
        if not self._problem_args and not self._problem_kwargs:
            return s + ")"
        for a in self._problem_args:
            val = str(a) if not isinstance(a, str) else "'%s'" % a
            s += val + ", "
        for k, v in self._problem_kwargs.items():
            val = str(v) if not isinstance(v, str) else "'%s'" % v
            s += str(k) + "=" + val + ", "
        return s[:-2] + ")"

    def __eq__(self, other):
        """__eq__.

        Determine if ``self`` and ``other`` define the same problem.

        Parameters
        ----------
        other : an object derived from the ``Problem`` class.

        Return
        ------
        eq : boolean.
            If ``self`` and ``other`` represent the same problem.

        """
        return (
            isinstance(other, type(self)) and
            self._problem_args == other._problem_args and
            self._problem_kwargs == other._problem_kwargs
        )

    def convert_solution(self, solution, *args, **kwargs):
        """convert_solution.

        Convert the solution to the QUBO to the solution to the problem.
        Should be implemented in child classes. If it is not implemented in the
        child class, then this function will by default return the same
        solution as what inputted.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool.
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        Implemented in the child class.

        """
        raise NotImplementedError("To be implemented in the child class")

    def is_solution_valid(self, solution, *args, **kwargs):
        """is_solution_valid.

        Returns whether or not the proposed solution is valid. Should be
        implemented in child classes. If it is not implemented in the child
        class, then this function will by default return True.

        Parameters
        ----------
        solution : iterable or dict.
            solution can be the output of ``convert_solution``,
            or the  QUBO or QUSO solver output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool.
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        valid : boolean.
            True if the proposed solution is valid, else False.

        """
        raise NotImplementedError("To be implemented in the child class")

    def solve_bruteforce(self, *args, **kwargs):
        """solve_bruteforce.

        Solve the problem bruteforce. THIS SHOULD NOT BE USED FOR LARGE
        PROBLEMS! This is implemented in the parent ``qubovert.utils.Problem``
        class. Some problems use a lot of slack binary variables for their
        QUBO/QUSO formulations. If this is the case, then the child class
        for this problem should override this method with a better bruteforce
        solver. But, for problems that do not use slack variables, this
        method will suffice. It converts the problem to QUBO, solves it with
        ``qubovert.utils.solve_qubo_bruteforce``, and then calls and returns
        ``convert_solution``.

        Parameters
        ----------
        aruments : arguments and keyword arguments.
            Contains args and kwargs for the ``to_qubo`` method. Also contains
            a ``all_solutions`` boolean flag, which indicates whether or not
            to return all the solutions, or just the best one found.
            ``all_solutions`` defaults to False.

        Return
        ------
        res : the output or outputs of the ``convert_solution`` method.
            If ``all_solutions`` is False, then ``res`` is just the output
            of the ``convert_solution`` method.
            If ``all_solutions`` is True, then ``res`` is a list of outputs
            of the ``convert_solution`` method, e.g. a converted solution
            for each solution that the bruteforce solver returns.

        """
        kwargs = kwargs.copy()
        all_solutions = kwargs.pop("all_solutions", False)
        qubo = self.to_qubo(*args, **kwargs)
        sol = qubo.solve_bruteforce(all_solutions)
        if all_solutions:
            return [self.convert_solution(x) for x in sol]
        return self.convert_solution(sol)

    def to_pubo(self, *args, **kwargs):
        """to_pubo.

        Since the model is already degree two, ``self.to_pubo`` will simply
        return ``qubovert.utils.PUBOMatrix(self.to_qubo(*args, **kwargs))``.

        Return
        ------
        P : qubovert.utils.PUBOMatrix object.
            The upper triangular PUBO matrix, a PUBOMatrix object.
            For most practical purposes, you can use PUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUBOMatrix)``.

        """
        return PUBOMatrix(self.to_qubo(*args, **kwargs))



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_pusomatrix.py.

This file contains the PUSOMatrix object.

"""

from . import (
    PUBOMatrix, hash_function,
    puso_value, solve_puso_bruteforce
)


__all__ = 'PUSOMatrix',


class PUSOMatrix(PUBOMatrix):
    """PUSOMatrix.

    ``PUSOMatrix`` inherits some methods from ``DictArithmetic``, see
    ``help(qubovert.utils.DictArithmetic)``.

    ``PUSOMatrix`` inherits some methods from ``PUBOMatrix``, see
    ``help(qubovert.utils.PUBOMatrix)``.

    A class to handle PUSO matrices. It is the same thing as a dictionary
    with some methods modified. Note that each key must be a tuple of integers
    >= 0.

    Note that below we only consider keys that are of length 1 or 2, but they
    can in general be arbitrarily long! See ``qubovert.utils.QUSOMatrix``
    for an object that restricts the length to <= 2.

    One method is that values will always default to 0. Consider the following
    example:

    >>> d = PUSOMatrix()
    >>> print(d[(0,)]) # will print 0
    >>> d[(0,)] += 1
    >>> print(d) # will print {(0,): 1}

    One method of PUSOMatrix is that it will always keep the PUSO
    upper triangular! Consider the following example:

    >>> d = PUSOMatrix()
    >>> d[(1, 0)] += 2
    >>> print(d)
    >>> # will print {(0, 1): 2}

    One method is that if we set an item to 0, it will be removed. Consider
    the following example:

    >>> d = PUSOMatrix()
    >>> d[(0,)] += 1
    >>> d[(0,)] -= 1
    >>> print(d) # will print {}

    One method is that if we initialize PUSOMatrix with a previous dictionary
    it will be reinitialized to ensure that the PUSOMatrix is upper
    triangular and contains no zero values. Consider the following example:

    >>> d = PUSOMatrix({(0,): 1, (1, 0): 2, (2, 0): 0, (2, 0, 1): 1})
    >>> print(d) # will print {(0,): 1, (0, 1): 2, (0, 1, 2): 1}

    We also change the update method so that it follows all the conventions.

    >>> d = PUSOMatrix({(0,): 1, (0, 1): 2})
    >>> d.update({(0,): 0, (1, 0): 1, (1, 2): -1})
    >>> print(d)  # will print {(0, 1): 1, (1, 2): -1}

    We also include arithmetic, addition, subtraction, scalar division,
    scalar multiplication, and all those in place. For example,

    >>> d = PUSOMatrix((0,)=1, (0, 1)=-2)
    >>> g = d + {(0,): -1}
    >>> print(g) # will print {(0, 1): -2}
    >>> g *= 4
    >>> print(g) # will print {(0, 1): -8}
    >>> g -= {(0, 1): -8}
    >>> print(g) # will print {}

    >>> d = PUSOMatrix({(0, 0): 1, (0, 1): -1})
    >>> print(d)
    {(): 1, (0, 1): -1}
    >>> g = {(0,): -1, (1,): 1}
    >>> d *= g
    >>> print(d)
    {(0,): -2, (1,): 2}

    >>> d = PUSOMatrix({(0, 0): 1, (0, 1): -1})
    >>> print(d ** 2 == d * d)
    True

    If we try to set a key with duplicated indices, it will squash the key
    into its equivalent form. For example ``(0, 1, 1, 2, 2, 2)`` is equivalent
    to ``(0, 2)``. Thus,

    >>> d = PUSOMatrix()
    >>> d[(0, 1, 1, 2, 2, 2)] += 1
    >>> d[(0, 2)] += 2
    >>> print(d)
    {(0, 2): 3}

    Adding or subtracting constants will update the () element of the
    dict.

    >>> d = PUSOMatrix()
    >>> d += 5
    >>> print(d)
    {(): 5}

    Finally, if you try to access a key out of order, it will sort the key. Be
    careful with this, it can cause unexpected behavior if you don't know it.
    For example,

    >>> d = PUSOMatrix()
    >>> d[(0, 1)] += 2
    >>> print(d[(1, 0)])  # will print 2

    """

    @classmethod
    def squash_key(cls, key):
        """squash_key.

        Will convert the input key into the standard form for PUSOMatrix /
        QUSOMatrix. It will get rid of pairs of duplicates and sort. This
        method will check to see if the input key is valid.

        Parameters
        ----------
        key : tuple of integers.

        Return
        ------
        k : tuple of integers.
            A sorted and squashed version of ``key``.

        Example
        -------
        >>> squash_key((0, 4, 0, 3, 3, 2, 3))
        >>> (2, 3, 4)

        """
        # if f is not None, then it is the squashed key. See QUSOMatrix.
        f = cls._check_key_valid(key)
        # here we use hash because some other classes that are subclasses of
        # this class will allow elements of the key to be strings! So we want
        # to still have something consistent to sort by. But for this class,
        # it doesn't make a difference, because hash_function(i) == i when i is
        # an int.
        return f or tuple(sorted(
            (x for x in set(key) if key.count(x) % 2),
            key=lambda x: hash_function(x)
        ))

    def solve_bruteforce(self, all_solutions=False):
        """solve_bruteforce.

        Solve the problem bruteforce. THIS SHOULD NOT BE USED FOR LARGE
        PROBLEMS! This is the exact same as calling
        ``qubovert.utils.solve_puso_bruteforce(self, all_solutions)[1]``.

        Parameters
        ----------
        all_solutions : bool.
            See the description of the ``all_solutions`` parameter in
            ``qubovert.utils.solve_puso_bruteforce``.

        Return
        ------
        res : the second element of the two element tuple that is returned from
            ``qubovert.utils.solve_puso_bruteforce``.

        """
        return solve_puso_bruteforce(self, all_solutions,
                                     self.is_solution_valid)[1]

    def value(self, x):
        r"""value.

        Find the value of
            :math:`\sum_{i,...,j} H_{i...j} z_{i} ... z_{j}`. Calling
            ``self.value(z)`` is the same as calling
            ``qubovert.utils.puso_value(z, self)``.

        Parameters
        ----------
        z: dict or iterable.
            Maps variable labels to their values, -1 or 1. Ie z[i] must be the
            value of variable i.

        Return
        ------
        value : float.
            The value of the PUSO/QUSO with the given assignment `z`.

        Example
        -------
        >>> from qubovert.utils import QUSOMatrix, PUSOMatrix
        >>> from qubovert import QUSO, PUSO

        >>> H = PUSOMatrix({(0, 1): -1, (0,): 1})
        >>> z = {0: -1, 1: 1}
        >>> H.value(z)
        0

        >>> H = PUSO({(0, 1): -1, (0,): 1})
        >>> z = {0: -1, 1: 1}
        >>> H.value(z)
        0

        >>> L = QUSOMatrix({(0, 1): -1, (0,): 1})
        >>> z = {0: -1, 1: 1}
        >>> L.value(z)
        0

        >>> L = QUSO({(0, 1): -1, (0,): 1})
        >>> z = {0: -1, 1: 1}
        >>> L.value(z)
        0

        """
        return puso_value(x, self)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_subgraph.py.

This file contains the function to create subgraphs and subvalues of QUBOs,
PUBOs, QUSOs, PUSOs, etc.

"""

import numpy as np


__all__ = 'subgraph', 'subvalue'


def subgraph(G, nodes, connections=None):
    """subgraph.

    Create the subgraph of ``G`` that only includes vertices in ``nodes``, and
    external nodes are given the values in ``connections``.

    Parameters
    ----------
    G : dict or any subclass of dict.
        ``G`` must contain tuple keys.
    nodes : set.
        Nodes of ``G`` to include in the subgraph.
    connections : dict (optional, defaults to {}).
        For each node in ``G`` that is not in ``nodes``, we assign a value
        given by ``connections.get(node, 0)``.

    Return
    ------
    D : same as type(G).
        The subgraph of ``G`` with nodes in ``nodes`` and the values of the
        nodes not included given by ``connections``.

    Notes
    -----
    Any offset value included in ``G`` (ie {(): 1}) will be ignored, however
    there may be an offset in the output ``D``.

    Examples
    --------
    >>> G = {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
    >>> D = subgraph(G, {0, 2}, {1: 5})
    >>> D
    {(0,): -17, (0, 2): -1, (): 10}

    >>> G = {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
    >>> D = subgraph(G, {0, 2})
    >>> D
    {(0, 2): -1, (0,): 3}

    >>> G = {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
    >>> D = subgraph(G, {0, 1}, {2: -10})
    >>> D
    {(0, 1): -4, (0,): 13, (1,): 2}

    >>> G = {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
    >>> D = subgraph(G, {0, 1})
    >>> D
    {(0, 1): -4, (0,): 3, (1,): 2}

    """
    if connections is None:
        connections = {}

    D = type(G)()
    for k, v in G.items():
        if not isinstance(k, tuple):
            raise ValueError("Keys must be tuples")
        if not k:
            continue
        key = tuple(filter(lambda x: x in nodes, k))
        not_key = filter(lambda x: x not in nodes, k)
        value = v * np.prod([connections.get(i, 0) for i in not_key])
        value += D.get(key, 0)
        if value:
            D[key] = value

    return D


def subvalue(values, G):
    """subvalue.

    Replace each element in ``G`` with a value in ``values`` if it exists.

    Parameters
    ----------
    values : dict.
        For each node ``v`` in ``G`` that is in ``values``, we replace the
        node with ``values[v]``.
    G : dict or any subclass of dict.
        ``G`` must contain tuple keys.

    Return
    ------
    D : same as type(G).

    Examples
    --------
    >>> G = {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
    >>> D = subvalue({0: 2}, G)
    >>> D
    {(1,): -6, (2,): -2, (): 8}

    >>> G = {(0, 1): -4, (0, 2): -1, (0,): 3, (1,): 2, (): 2}
    >>> D = subvalue({2: -3}, G)
    >>> D
    {(0, 1): -4, (0,): 6, (1,): 2, (): 2}

    """
    D = type(G)()
    for k, v in G.items():
        if not isinstance(k, tuple):
            raise ValueError("Keys must be tuples")
        key = tuple(filter(lambda x: x not in values, k))
        not_key = filter(lambda x: x in values, k)
        value = v * np.prod([values[i] for i in not_key])
        value += D.get(key, 0)
        if value:
            D[key] = value

    return D



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_vertex_cover.py.

Contains the VertexCover class. See ``help(qubovert.problems.VertexCover)``.

"""

from qubovert import PCBO
from qubovert.problems import Problem
from qubovert.utils import (
    solution_type, spin_to_boolean, QUBOMatrix, hash_function
)


__all__ = 'VertexCover',


class VertexCover(Problem):
    """VertexCover.

    Class to manage converting Vertex Cover to and from its QUBO and
    QUSO formluations. Based on the paper "Ising formulations of many
    NP problems", hereforth designated as [Lucas].

    The goal of the VertexCover problem is to find the smallest number of
    verticies that can be colored such that every edge of the graph is
    incident to a colored vertex.

    VertexCover inherits some methods and attributes from the Problem class.
    See ``help(qubovert.problems.Problem)``.

    Example
    -------
    >>> from qubovert.problems import VertexCover
    >>> from any_module import qubo_solver
    >>> # or you can use my bruteforce solver...
    >>> # from qubovert.utils import solve_qubo_bruteforce as qubo_solver
    >>> edges = {("a", "b"), ("a", "c"), ("c", "d"), ("a", "d")}
    >>> problem = VertexCover(edges)
    >>> Q  = problem.to_qubo()
    >>> obj, sol = qubo_solver(Q)
    >>> solution = problem.convert_solution(sol)

    >>> print(solution)
    {"a", "c"}

    >>> print(problem.is_solution_valid(solution))
    True  # since each edge is adjacent to either "a" or "c".

    >>> print(obj == len(solution))
    True

    """

    def __init__(self, edges):
        """__init__.

        The goal of the VertexCover problem is to find the smallest number of
        verticies that can be coloredsuch that every edge of the graph is
        incident to a colored vertex.  All naming conventions follow the names
        in the paper [Lucas].

        Parameters
        ----------
        edges : set of two element tuples.
            Describes edges of the graph.

        Examples
        --------
        >>> edges = {("a", "b"), ("a", "c")}
        >>> problem = VertexCover(edges)

        >>> edges = {(0, 1), (0, 2)}
        >>> problem = VertexCover(edges)

        """
        self._edges = edges.copy()
        self._vertices = {y for x in edges for y in x}
        self._vertex_to_index = {
            x: i
            for i, x in enumerate(sorted(self._vertices, key=hash_function))
        }
        self._index_to_vertex = {
            i: x for x, i in self._vertex_to_index.items()
        }
        self._N, self._n = len(self._vertices), len(self._edges)

    @property
    def E(self):
        """E.

        A copy of the set of edges of the graph. Updating the copy will not
        update the instance set.

        Return
        ------
        E : set of two element tuples.
            A copy of the edge set defining the Vertex Cover problem.

        """
        return self._edges.copy()

    @property
    def V(self):
        """V.

        A copy of the vertex set of the graph. Updating the copy will not
        update the instance set.

        Return
        ------
        V : set.
            A copy of the set of vertices corresponding to the edge set for the
            Vertex Cover problem.

        """
        return self._vertices.copy()

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        The number of binary variables that the QUBO and QUSO use.

        Return
        ------
        num : integer.
            The number of variables in the QUBO/QUSO formulation.

        """
        return self._N

    def to_qubo(self, A=2, B=1):
        r"""to_qubo.

        Create and return the vertex cover problem in QUBO form following
        section 4.3 of [Lucas]. The Q matrix for the QUBO
        will be returned as an uppertriangular dictionary. Thus, the problem
        becomes minimizing :math:`\sum_{i \leq j} x_i x_j Q_{ij}`. ``A`` and
        ``B`` are parameters to enforce constraints.

        It is formatted such that if all the constraints are satisfied, then
        the objective function will be equal to the total number of colored
        verticies.

        Parameters
        ----------
        A: positive float (optional, defaults to 2).
            A enforces the constraints. See section 4.3 of [Lucas].
        B: positive float that is less than A (optional, defaults to 1).
            See section 4.3 of [Lucas].

        Return
        ------
        Q : qubovert.utils.QUBOMatrix object.
            The upper triangular QUBO matrix, a QUBOMatrix object.
            For most practical purposes, you can use QUBOMatrix in the
            same way as an ordinary dictionary. For more information,
            see help(qubovert.utils.QUBOMatrix).

        Example
        -------
        >>> problem = VertexCover({(0, 1), (0, 2)})
        >>> Q = problem.to_qubo()

        """
        # all naming conventions follow the paper listed in the docstring

        Q = QUBOMatrix()

        # encode H_B (equation 34)
        for i in range(self._N):
            Q[(i,)] += B

        # encode H_A, ie each edge is adjacent to at least one colored vertex.
        for u, v in self._edges:
            iu, iv = self._vertex_to_index[u], self._vertex_to_index[v]
            Q += PCBO().add_constraint_OR(iu, iv, lam=A)

        return Q

        # this is slower, since we create a PCBO and then a QUBOMatrix
        # H = PCBO()
        # H.set_mapping(self._vertex_to_index)

        # # encode H_B (equation 34)
        # for v in self._vertices:
        #     H[(v,)] += B

        # # encode H_A, ie each edge is adjacent to at least one colored vertex
        # for u, v in self._edges:
        #     H.add_constraint_OR(u, v, lam=A)

        # return H.to_qubo()

    def convert_solution(self, solution, spin=False):
        """convert_solution.

        Convert the solution to the QUBO or QUSO to the solution to the Vertex
        Cover problem.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        res : set.
            A set of which verticies need to be colored. Thus, if this
            function returns {0, 2}, then this means that vertex 0 and 2
            should be colored.

        """
        if not isinstance(solution, dict):
            solution = dict(enumerate(solution))
        sol_type = solution_type(solution, 'spin' if spin else 'bool')
        if sol_type == 'spin':
            solution = spin_to_boolean(solution)
        return set(
            self._index_to_vertex[i] for i, x in solution.items() if x
        )

    def is_solution_valid(self, solution, spin=False):
        """is_solution_valid.

        Returns whether or not the proposed solution satisfies the constraint
        that every edge has at least one colored vertex.

        Parameters
        ----------
        solution : iterable or dict.
            solution can be the output of VertexCover.convert_solution,
            or the  QUBO or QUSO solver output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        valid : boolean.
            True if the proposed solution is valid, else False.

        """
        if not isinstance(solution, set):
            solution = self.convert_solution(solution, spin)

        return all(i in solution or j in solution for i, j in self._edges)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""``benchmarking`` contains many benchmarking example problems.

This module contains many benchmarking problems. We import all the problems
globally for user use. See the ``__all__`` value for all the problems imported.

"""

from ._alternating_sectors_chain import *

from ._alternating_sectors_chain import __all__ as __all_asc__

__all__ = (
    __all_asc__
)

name = "benchmarking"



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_bo_parentclass.py.

Contains the BO parent class. See ``help(qubovert.utils.BO)``. Used as a parent
class for some Binary Optimization classes, such as ``qubovert.QUBO``,
``qubovert.QUSO``, etc.

"""

from . import Conversions
from ._dict_arithmetic import _generate_key_value_pairs


__all__ = 'BO',


class BO(Conversions):
    """BO.

    Parent class for some Binary Optimization classes, such as
    ``qubovert.QUBO``, ``qubovert.QUSO``, etc.

    BO inherits some methods and attributes the ``DictArithmetic`` class.
    See ``help(qubovert.utils.DictArithmetic)``.

    BO inherits some methods and attributes the ``Conversions`` class.
    See ``help(qubovert.utils.Conversions)``.

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with Binary Optimization models. See child
        classes for info on inputs.

        Parameters
        ----------
        arguments : Defined in child classes.

        """
        self._mapping, self._reverse_mapping, self._next_label = {}, {}, 0

    @property
    def mapping(self):
        """mapping.

        Return a copy of the mapping dictionary that maps the provided
        labels to integers from 0 to n-1, where n is the number of variables
        in the problem.

        Return
        ------
        mapping : dict.
            Dictionary that maps provided labels to integer labels.

        """
        return self._mapping.copy()

    @property
    def reverse_mapping(self):
        """reverse_mapping.

        Return a copy of the reverse_mapping dictionary that maps the integer
        labels to the provided labels. Opposite of ``mapping``.

        Return
        ------
        reverse_mapping : dict.
            Dictionary that maps integer labels to provided labels.

        """
        return self._reverse_mapping.copy()

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        Return the number of binary variables in the problem.

        Return
        ------
        n : int.
            Number of binary variables in the problem.

        """
        return self._next_label

    @property
    def max_index(self):
        """max_index.

        Return the maximum label of the integer labeled version of the problem.

        Return
        ------
        m : int.

        """
        return self.num_binary_variables - 1

    def set_mapping(self, *args, **kwargs):
        """set_mapping.

        ``BO`` sublcasses automatically create a mapping from variable names to
        integers as they are being built. However, the mapping is based on the
        order in which elements are entered and therefore may not be as
        desired. Of course, the ``convert_solution`` method keeps track of the
        mapping and can/should always be used. But if you want a consistent
        mapping, then ``set_mapping`` can be used.

        Consider the following examples (we use the ``qubovert.QUBO`` class for
        the examples, which is a subclass of ``BO``).

        Example 1:

        >>> from qubovert import QUBO
        >>> Q = QUBO()
        >>> Q[(0,)] += 1
        >>> Q[(1,)] += 2
        >>> Q.mapping
        {0: 0, 1: 1}
        >>> Q.to_qubo()
        {(0,): 1, (1,): 2}

        Example 2:

        >>> from qubovert import QUBO
        >>> Q = QUBO()
        >>> Q[(1,)] += 2
        >>> Q[(0,)] += 1
        >>> Q.mapping
        {0: 1, 1: 0}
        >>> Q.to_qubo()
        {(0,): 2, (1,): 1}

        To ensure consistency in mappings, you can provide your own mapping
        with ``set_mapping``. See the following modified examples.

        Modified example 1:

        >>> from qubovert import QUBO
        >>> Q = QUBO()
        >>> Q[(0,)] += 1
        >>> Q[(1,)] += 2
        >>> Q.set_mapping({0: 0, 1: 1})
        >>> Q.mapping
        {0: 0, 1: 1}
        >>> Q.to_qubo()
        {(0,): 1, (1,): 2}

        Modified example 2:

        >>> from qubovert import QUBO
        >>> Q = QUBO()
        >>> Q[(1,)] += 2
        >>> Q[(0,)] += 1
        >>> Q.set_mapping({0: 0, 1: 1})
        >>> Q.mapping
        {0: 0, 1: 1}
        >>> Q.to_qubo()
        {(0,): 1, (1,): 2}

        Parameters
        ----------
        arguments : defines a dictionary with ``d = dict(*args, **kwargs)``.
            ``d`` will become the mapping. See ``help(self.mapping)``

        Notes
        -----
        Using ``set_mapping`` to set the mapping will also automatically
        set the ``reverse_mapping``, so there is no need to call both
        ``set_mapping`` and ``set_reverse_mapping``.

        """
        self._mapping, self._reverse_mapping = {}, {}
        for k, v in _generate_key_value_pairs(*args, **kwargs):
            self._mapping[k] = v
            self._reverse_mapping[v] = k

    def set_reverse_mapping(self, *args, **kwargs):
        """set_reverse_mapping.

        Same as ``set_mapping`` but reversed. See
        ``help(self.reverse_mapping)`` and ``help(self.set_mapping)``.

        Parameters
        ----------
        arguments : defines a dictionary with ``d = dict(*args, **kwargs)``.
            ``d`` will become the reverse mapping. See
            ``help(self.reverse_mapping)``.

        Notes
        -----
        Using ``set_reverse_mapping`` to set the mapping will also
        automatically set the ``mapping``, so there is no need to call both
        ``set_mapping`` and ``set_reverse_mapping``.

        """
        self._mapping, self._reverse_mapping = {}, {}
        for k, v in _generate_key_value_pairs(*args, **kwargs):
            self._mapping[v] = k
            self._reverse_mapping[k] = v

    def __setitem__(self, key, value):
        """__setitem__.

        Overrides the dict.__setitem__ command. If `value` is equal to 0, then
        the key will be removed from the dictionary. Thus no elements in the
        QUSO dictionary will ever have zero value.

        Parameters
        ----------
        key : tuple.
            Element of the dictionary that follows convensions of the class.
        value : numeric.
            Value corresponding to the key.

        """
        super().__setitem__(key, value)

        for i in key:
            if i not in self._mapping:
                self._mapping[i] = self._next_label
                self._reverse_mapping[self._next_label] = i
                self._next_label += 1



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_quso.py.

Contains the QUSO class. See ``help(qubovert.QUSO)``.

"""

from .utils import (
    BO, QUSOMatrix, PUSOMatrix, solution_type, boolean_to_spin
)


__all__ = 'QUSO',


class QUSO(BO, QUSOMatrix):
    """QUSO.

    Class to manage converting general QUSO problems to and from their
    QUBO and QUSO formluations.

    This class deals with QUSOs that have spin labels that do not range from
    0 to n-1. If your labels are nonnegative integers, consider using
    ``qubovert.utils.QUSOMatrix``. Note that it is generally
    more efficient to initialize an empty QUSO object and then build the
    QUSO, rather than initialize a QUSO object with an already built dict.

    QUSO inherits some methods and attributes the ``QUSOMatrix`` class. See
    ``help(qubovert.utils.QUSOMatrix)``.

    QUSO inherits some methods and attributes the ``BO`` class. See
    ``help(qubovert.utils.BO)``.

    Example
    -------
    >>> quso = QUSO()
    >>> quso[('a',)] += 5
    >>> quso[(0, 'a')] -= 2
    >>> quso -= 1.5
    >>> quso
    {('a',): 5, ('a', 0): -2, (): -1.5}

    >>> quso = QUSO({('a',): 5, (0, 'a'): -2, (): -1.5})
    >>> quso
    {('a',): 5, ('a', 0): -2, (): -1.5}
    >>> L = quso.to_quso()
    >>> L
    {(0,): 5, (0, 1): -2, (): -1.5}
    >>> quso.convert_solution({0: 1, 1: 0})
    {'a': 1, 0: 0}

    Note
    ----
    Note that keys will end up sorted by their hash. Hashes will not be
    consistent across Python sessions (unless they are integers)! For example,
    both of the following can happen:

    >>> print(QUSO({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (0, 1): -1}

    or

    >>> print(QUSO({('a', 0): 1, (0, 1): -1}))
    {(0, 'a'): 1, (0, 1): -1}

    But the following will never happen:

    >>> print(QUSO({('a', 0): 1, (0, 1): -1}))
    {('a', 0): 1, (1, 0): -1}

    Ie integers will always be correctly sorted.

    Note
    ----
    For efficiency, many internal variables including mappings are computed as
    the problemis being built. This can cause these
    values to be wrong for some specific situations. Calling ``refresh``
    will rebuild the dictionary, resetting all of the values. See
    ``help(QUSO.refresh)``

    Examples
    --------
    >>> from qubovert import QUSO
    >>> L = QUSO()
    >>> L[('a',)] += 1
    >>> L, L.mapping, L.reverse_mapping
    {('a',): 1}, {'a': 0}, {0: 'a'}
    >>> L[('a',)] -= 1
    >>> L, L.mapping, L.reverse_mapping
    {}, {'a': 0}, {0: 'a'}
    >>> L.refresh()
    >>> L, L.mapping, L.reverse_mapping
    {}, {}, {}

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with QUSOs that have spin labels that do not range
        from 0 to n-1. If your labels are nonnegative integers, consider using
        ``qubovert.utils.QUSOMatrix``. Note that it is generally more
        efficient to initialize an empty QUSO object and then build the QUSO,
        rather than initialize a QUSO object with an already built dict.

        Parameters
        ----------
        argguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class.

        Examples
        --------
        >>> quso = QUSO()
        >>> quso[('a',)] += 5
        >>> quso[(0, 'a')] -= 2
        >>> quso -= 1.5
        >>> quso
        {('a',): 5, ('a', 0): -2, (): -1.5}

        >>> quso = QUSO({('a',): 5, (0, 'a'): -2, (): -1.5})
        >>> quso
        {('a',): 5, ('a', 0): -2, (): -1.5}

        """
        BO.__init__(self, *args, **kwargs)
        QUSOMatrix.__init__(self, *args, **kwargs)

    def to_quso(self):
        """to_quso.

        Create and return upper triangular QUSO representing the problem.
        The labels will be integers from 0 to n-1.

        Return
        ------
        L : qubovert.utils.QUSOMatrix object.
            The upper triangular QUSO matrix, a QUSOMatrix object.
            For most practical purposes, you can use QUSOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.QUSOMatrix)``.

        """
        L = QUSOMatrix()

        for k, v in self.items():
            key = tuple(self._mapping[i] for i in k)
            L[key] += v

        return L

    def to_puso(self):
        """to_puso.

        Since the model is already a QUSO, ``self.to_puso`` will simply
        return ``qubovert.utils.PUSOMatrix(self.to_quso())``.

        Return
        ------
        H : qubovert.utils.PUSOMatrix object.
            The upper triangular PUSO matrix, a PUSOMatrix object.
            For most practical purposes, you can use PUSOMatrix in the
            same way as an ordinary dictionary. For more information,
            see ``help(qubovert.utils.PUSOMatrix)``.

        """
        return PUSOMatrix(self.to_quso())

    def convert_solution(self, solution, spin=True):
        """convert_solution.

        Convert the solution to the integer labeled QUSO to the solution to
        the originally labeled QUSO.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUSO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to True).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        res : dict.
            Maps spin variable labels to their QUSO solutions values {1, -1}.

        Example
        -------
        >>> quso = QUSO({('a',): 1, ('a', 'b'): -2, ('c',): -1})
        >>> L = quso.to_quso()
        >>> L
        {(0,): 1, (0, 1): -2, (2,): -1}
        >>> solution = solve_quso(L)  # any solver you want
        >>> solution
        [1, 1, -1]  # or {0: 1, 1: 1, 2: -1}
        >>> sol = quso.convert_solution(solution)
        >>> sol
        {'a': 1, 'b': 1, 'c': -1}

        >>> quso = QUSO({('a',): 1, ('a', 'b'): -2, ('c',): -1})
        >>> L = quso.to_qubo()
        >>> solution = solve_quso(L)  # any solver you want
        >>> solution
        [0, 0, 1]  # or {0: 0, 1: 0, 2: 1}
        >>> sol = quso.convert_solution(solution)
        >>> sol
        {'a': 1, 'b': 1, 'c': -1}

        """
        sol_type = solution_type(solution, 'spin' if spin else 'bool')
        if sol_type == 'bool':
            solution = boolean_to_spin(solution)
        return {
            self._reverse_mapping[i]: solution[i]
            for i in range(self.num_binary_variables)
        }

    @staticmethod
    def _check_key_valid(key):
        """_check_key_valid.

        Internal method to check if an input key to the dictionary is valid.
        Checks to see if ``key`` is a tuple.

        Parameters
        ----------
        key : anything, but must be a tuple to be valid.

        Returns
        -------
        None.

        Raises
        ------
        KeyError if the key is invalid.

        """
        # override QUSOMatrix._check_key_valid to allow for noninteger keys.
        invalid = (
            not isinstance(key, tuple) or
            len(set(x for x in set(key) if key.count(x) % 2)) > 2
        )
        if invalid:
            raise KeyError(
                "Key formatted incorrectly, must be tuple of <= 2 unique "
                "element. See PUSO for arbitrary number of unique elements.")



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_pcso.py.

Contains the PCSO class. See ``help(qubovert.PCSO)``.

"""

from . import PUSO, PCBO
from .utils import puso_to_pubo, pubo_to_puso


__all__ = 'PCSO', 'spin_var'


def spin_var(name):
    """spin_var.

    Create a PCSO (see ``qubovert.PCSO``) from a single spin variable.

    Parameters
    ----------
    name : any hashable object.
        Name of the spin variable.

    Return
    ------
    pcso : qubovert.PCSO object.
        The model representing the spin variable.

    Examples
    --------
    >>> from qubovert import spin_var, PCSO
    >>>
    >>> z0 = spin_var("z0")
    >>> print(z0)
    {('z0',): 1}
    >>> print(isinstance(z0, PCSO))
    True
    >>> print(z0.name)
    z0

    >>> z = [spin_var('z{}'.format(i)) for i in range(5)]
    >>> pcso = sum(z)
    >>> print(pcso)
    {('z0',): 1, ('z1',): 1, ('z2',): 1, ('z3',): 1, ('z4',): 1}
    >>> pcso **= 2
    >>> print(pcso)
    {(): 5, ('z0', 'z1'): 2, ('z2', 'z0'): 2, ('z3', 'z0'): 2, ('z0', 'z4'): 2,
     ('z2', 'z1'): 2, ('z3', 'z1'): 2, ('z4', 'z1'): 2, ('z3', 'z2'): 2,
     ('z2', 'z4'): 2, ('z3', 'z4'): 2}
    >>> pcso *= -1
    >>> print(pcso.solve_bruteforce(all_solutions=True))
    [{'z0': -1, 'z1': -1, 'z2': -1, 'z3': -1, 'z4': -1},
     {'z0': 1, 'z1': 1, 'z2': 1, 'z3': 1, 'z4': 1}]
    >>> pcso.add_constraint_eq_zero(z[0] + z[1])
    >>> print(pcso.solve_bruteforce(all_solutions=True))
    [{'z0': -1, 'z1': 1, 'z2': -1, 'z3': -1, 'z4': -1},
     {'z0': -1, 'z1': 1, 'z2': 1, 'z3': 1, 'z4': 1},
     {'z0': 1, 'z1': -1, 'z2': -1, 'z3': -1, 'z4': -1},
     {'z0': 1, 'z1': -1, 'z2': 1, 'z3': 1, 'z4': 1}]

    """
    pcso = PCSO({(name,): 1})
    pcso.name = name
    return pcso


def _empty_pcbo(pcso):
    """_empty_pcbo.

    Create an empty PCBO whose ancilla variables begin at
    ``pcso.num_ancillas``.

    Parameters
    ----------
    pcso : PCSO object.

    Return
    ------
    pcbo : PCBO object.

    """
    h = PCBO()
    h._ancilla = pcso._ancilla
    return h


class PCSO(PUSO):
    """PCBO.

    This class deals with Polynomial Constrained Spin Optimization. PCSO
    inherits some methods and attributes from the ``PUSO`` class. See
    ``help(qubovert.PUSO)``.

    ``PCSO`` has all the same methods as ``PUSO``, but adds some constraint
    methods; namely

    - ``add_constraint_eq_zero(H, lam=1, ...)`` enforces that ``H == 0`` by
      penalizing with ``lam``,
    - ``add_constraint_ne_zero(H, lam=1, ...)`` enforces that ``H != 0`` by
      penalizing with ``lam``,
    - ``add_constraint_lt_zero(H, lam=1, ...)`` enforces that ``H < 0`` by
      penalizing with ``lam``,
    - ``add_constraint_le_zero(H, lam=1, ...)`` enforces that ``H <= 0`` by
      penalizing with ``lam``,
    - ``add_constraint_gt_zero(H, lam=1, ...)`` enforces that ``H > 0`` by
      penalizing with ``lam``, and
    - ``add_constraint_ge_zero(H, lam=1, ...)`` enforces that ``H >= 0`` by
      penalizing with ``lam``.

    Each of these takes in a PUSO ``H`` and a lagrange multiplier ``lam``
    that defaults to 1. See each of their docstrings for important details on
    their implementation.

    Notes
    -----
    - Variables names that begin with ``"__a"`` should not be used since they
      are used internally to deal with some ancilla variables to enforce
      constraints.
    - The ``self.solve_bruteforce`` method will solve the PCSO ensuring that
      all the inputted constraints are satisfied. Whereas
      ``qubovert.utils.solve_puso_bruteforce(self)`` or
      ``qubovert.utils.solve_puso_bruteforce(self.to_pubo())`` will solve the
      PUSO created from the PCSO. If the inputted constraints are not
      enforced strong enough (ie too small lagrange multipliers) then these may
      not give the correct result, whereas ``self.solve_bruteforce()`` will
      always give the correct result (ie one that satisfies all the
      constraints).

    Examples
    --------
    See ``qubovert.PUSO`` for more examples of using PCSO without
    constraints. See ``qubovert.PCBO`` for many constraint examples in PUBO
    form. ``PCSO`` is the same but converting to PUSO.

    >>> H = PCSO()
    >>> H.add_constraint_eq_zero({('a', 1): 2, (1, 2): -1, (): -1})
    >>> H
    {(): 6, ('a', 2): -4, ('a', 1): -4, (1, 2): 2}

    >>> H = PCSO()
    >>> H.add_constraint_eq_zero(
            {(0, 1): 1}
        ).add_constraint_lt_zero(
            {(1, 2): 1, (): -1}
        )

    """

    def __init__(self, *args, **kwargs):
        """__init__.

        This class deals with polynomial constrained spin optimization.
        Note that it is generally more efficient to initialize an empty PCSO
        object and then build the PCSO, rather than initialize a PCSO object
        with an already built dict.

        Parameters
        ----------
        arguments : define a dictionary with ``dict(*args, **kwargs)``.
            The dictionary will be initialized to follow all the convensions of
            the class. Alternatively, ``args[0]`` can be a PCSO.

        Examples
        --------
        >>> pcso = PCSO()
        >>> pcso[('a',)] += 5
        >>> pcso[(0, 'a')] -= 2
        >>> pcso -= 1.5
        >>> pcso
        {('a',): 5, ('a', 0): -2, (): -1.5}
        >>> pcso.add_constraint_eq_zero({('a',): 1, ('b',): 1}, lam=5)
        >>> pcso
        {('a',): 5, ('a', 0): -2, (): 8.5, ('a', 'b'): 10}

        >>> pcso = PCSO({('a',): 5, (0, 'a', 1): -2, (): -1.5})
        >>> pcso
        {('a',): 5, ('a', 0, 1): -2, (): -1.5}

        """
        PCBO.__init__(self, *args, **kwargs)

    def update(self, *args, **kwargs):
        """update.

        Update the PCSO but following all the conventions of this class.

        Parameters
        ----------
        ``*args/**kwargs`` : defines a dictionary or PCSO.
            Ie ``d = dict(*args, **kwargs)``.
            Each element in d will be added in place to this instance following
            all the required convensions.

        """
        PCBO.update(self, *args, **kwargs)

    @property
    def constraints(self):
        """constraints.

        Return the constraints of the PCSO.

        Return
        ------
        res : dict.
            The keys of ``res`` are ``'eq'``, ``'ne'``, ``'lt'``, ``'le'``,
            ``'gt'``, and ``'ge'``. The values are lists of ``qubovert.PUSO``
            objects. For a given key, value pair ``k, v``, the ``v[i]`` element
            represents the PUSO ``v[i]`` being
            == 0 if ``k == 'eq'``, != 0 if ``k == 'ne'``,
            < 0 if ``k == 'lt'``, <= 0 if ``k == 'le'``,
            > 0 if ``k == 'gt'``, >= 0 if ``k == 'ge'``.

        """
        return {k: [x.copy() for x in v] for k, v in self._constraints.items()}

    def _append_constraint(self, key, constraint):
        """_append_constraint.

        Internal method to add a constraint to the constraints dictionary.

        Parameters
        ----------
        key : str.
            One of ``'eq'``, ``'lt'``, ``'le'``, ``'gt'``, or ``'ge'``.
        constraint : qubovert.PUSO object.

        """
        PCBO._append_constraint(self, key, constraint)

    @property
    def num_ancillas(self):
        """num_ancillas.

        Return the number of ancilla variables introduced to the PCSO in
        order to enforce the inputted constraints.

        Returns
        -------
        num : int.
            Number of ancillas in the PCSO.

        """
        return self._ancilla

    @classmethod
    def remove_ancilla_from_solution(cls, solution):
        """remove_ancilla_from_solution.

        Take a solution to the PCSO and remove all the ancilla variables, (
        represented by `_a` prefixes).

        Parameters
        ----------
        solution : dict.
            Must be the solution in terms of the original variables. Thus if
            ``solution`` is the solution to the ``self.to_pubo``,
            ``self.to_qubo``, ``self.to_puso``, or ``self.to_quso``
            formulations, then you should first call ``self.convert_solution``.
            See ``help(self.convert_solution)``.

        Return
        ------
        res : dict.
            The same as ``solution`` but with all the ancilla bits removed.

        """
        return PCBO.remove_ancilla_from_solution(solution)

    def is_solution_valid(self, solution):
        """is_solution_valid.

        Finds whether or not the given solution satisfies the constraints.

        Parameters
        ----------
        solution : dict.
            Must be the solution in terms of the original variables. Thus if
            ``solution`` is the solution to the ``self.to_pubo``,
            ``self.to_qubo``, ``self.to_puso``, or ``self.to_quso``
            formulations, then you should first call ``self.convert_solution``.
            See ``help(self.convert_solution)``.

        Return
        ------
        valid : bool.
            Whether or not the given solution satisfies the constraints.

        """
        return PCBO.is_solution_valid(self, solution)

    # override
    def __round__(self, ndigits=None):
        """round.

        Round values of the PCSO object.

        Parameters
        ----------
        ndigits : int.
            Number of decimal digits to round to.

        Returns
        -------
        res : PCSO object.
            Copy of self but with each value rounded to ``ndigits`` decimal
            digits. Each value has a type according to the docstring
            specifications of ``round``, see ``help(round)``.

        """
        return PCBO.__round__(self, ndigits)

    # override
    def subs(self, *args, **kwargs):
        """subs.

        Replace any ``sympy`` symbols that are used in the dict with values.
        Please see ``help(sympy.Symbol.subs)`` for more info.

        Parameters
        ----------
        arguments : substitutions.
            Same parameters as are inputted into ``sympy.Symbol.subs``.

        Returns
        -------
        res : a PCSO object.
            Same as ``self`` but with all the symbols replaced with values.

        """
        return PCBO.subs(self, *args, **kwargs)

    def add_constraint_eq_zero(self,
                               H, lam=1,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_eq_zero.

        Enforce that ``H == 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        H : dict representing a PUSO.
            The PUSO constraint such that H == 0. Note that ``H`` will be
            converted to a ``qubovert.PUSO`` object if it is not already,
            thus it must follow the conventions, see ``help(qubovert.PUSO)``.
            Please note that if ``H`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUSO ``H`` can take. If ``bounds`` is None, then they may be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCSO.
            Updates the PCSO in place, but returns ``self`` so that operations
            can be strung together.

        Examples
        --------
        The following enforces that :math:`\sum_{i=1}^{3} i z_i z_{i+1} == 0`.

        >>> H = PCSO()
        >>> H.add_constraint_eq_zero({(1, 2): 1, (2, 3): 2, (3, 4): 3})

        Here we show how operations can be strung together.

        >>> H = PCSO()
        >>> H.add_constraint_lt_zero(
                {(0, 1): 1}
            ).add_constraint_eq_zero(
                {(1, 2): 1, (): -1}
            )

        """
        H = PUSO(H)
        self._append_constraint("eq", H)

        h = _empty_pcbo(self).add_constraint_eq_zero(
            puso_to_pubo(H), lam=lam,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._ancilla = h._ancilla
        self += pubo_to_puso(h)
        return self

    def add_constraint_ne_zero(self,
                               H, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_ne_zero.

        Enforce that ``H != 0`` by penalizing invalid solutions with ``lam``.
        See Notes below for more details.

        Parameters
        ----------
        H : dict representing a PUSO.
            The PUSO constraint such that H != 0. Note that ``H`` will be
            converted to a ``qubovert.PUSO`` object if it is not already,
            thus it must follow the conventions, see ``help(qubovert.PUSO)``.
            Please note that if ``H`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUSO ``H`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCSO.
            Updates the PCSO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints. However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function.
        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\min_x \text{P.value(x)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\min_x \text{P.value(x)}|` ancilla
          bits will be used.

        """
        H = PUSO(H)
        self._append_constraint("ne", H)
        h = _empty_pcbo(self).add_constraint_ne_zero(
            puso_to_pubo(H), lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._ancilla = h._ancilla
        self += pubo_to_puso(h)
        return self

    def add_constraint_lt_zero(self,
                               H, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_lt_zero.

        Enforce that ``H < 0`` by penalizing invalid solutions with ``lam``.
        See Notes below for more details.

        Parameters
        ----------
        H : dict representing a PUSO.
            The PUSO constraint such that H < 0. Note that ``H`` will be
            converted to a ``qubovert.PUSO`` object if it is not already,
            thus it must follow the conventions, see ``help(qubovert.PUSO)``.
            Please note that if ``H`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUSO ``H`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCSO.
            Updates the PCSO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`z_0 + 2z_1 + ... < 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = PCSO()
          >>> H.add_constraint_lt_zero(
          >>>     {(0,): 0.5, (): 1.65, (1,): 1.0, (2,): -0.25})
          >>> test_sol = {0: -1, 1: -1, 2: 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: -1, 1: -1, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\min_z \text{H.value(z)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\min_z \text{H.value(z)}|` ancilla
          bits will be used.

        """
        H = PUSO(H)
        self._append_constraint("lt", H)
        h = _empty_pcbo(self).add_constraint_lt_zero(
            puso_to_pubo(H), lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._ancilla = h._ancilla
        self += pubo_to_puso(h)
        return self

    def add_constraint_le_zero(self,
                               H, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_le_zero.

        Enforce that ``H <= 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        H : dict representing a PUSO.
            The PUSO constraint such that H <= 0. Note that ``H`` will be
            converted to a ``qubovert.PUSO`` object if it is not already,
            thus it must follow the conventions, see ``help(qubovert.PUSO)``.
            Please note that if ``H`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUSO ``H`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCSO.
            Updates the PCSO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`z_0 + 2z_1 + ... \leq 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = PCSO()
          >>> H.add_constraint_le_zero(
                  {(0,): 0.5, (): 1.15, (1,): 1.0, (2,): -0.75})
          >> H
          {(0,): 1.65, (): 4.785, (0, 1): 1.0, (1,): 3.3, (0, 2): -0.75,
           (2,): -2.4749999999999996, ('__a0', 0): 0.5, ('__a0',): 1.65,
           (1, 2): -1.5, ('__a0', 1): 1.0, ('__a0', 2): -0.75}
          >>> test_sol = {0: -1, 1: -1, 2: 1, '__a0': 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: -1, 1: -1, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\min_z \text{H.value(z)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\min_z \text{H.value(z)}|` ancilla
          bits will be used.

        """
        H = PUSO(H)
        self._append_constraint("le", H)
        h = _empty_pcbo(self).add_constraint_le_zero(
            puso_to_pubo(H), lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._ancilla = h._ancilla
        self += pubo_to_puso(h)
        return self

    def add_constraint_gt_zero(self,
                               H, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_gt_zero.

        Enforce that ``H > 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        H : dict representing a PUSO.
            The PUSO constraint such that H > 0. Note that ``H`` will be
            converted to a ``qubovert.PUSO`` object if it is not already,
            thus it must follow the conventions, see ``help(qubovert.PUSO)``.
            Please note that if ``H`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUSO ``H`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCSO.
            Updates the PCSO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`z_0 + 2z_1 + ... > 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = PCSO()
          >>> H.add_constraint_gt_zero(
                  {(0,): -0.5, (): -1.65, (1,): -1.0, (2,): 0.25})
          >>> test_sol = {0: -1, 1: -1, 2: 1}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(sol)
          0.01

          {0: -1, 1: -1, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\max_z \text{H.value(z)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\max_z \text{H.value(z)}|` ancilla
          bits will be used.

        """
        H = PUSO(H)
        self._append_constraint("gt", H)
        h = _empty_pcbo(self).add_constraint_gt_zero(
            puso_to_pubo(H), lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._ancilla = h._ancilla
        self += pubo_to_puso(h)
        return self

    def add_constraint_ge_zero(self,
                               H, lam=1, log_trick=True,
                               bounds=None, suppress_warnings=False):
        r"""add_constraint_ge_zero.

        Enforce that ``H >= 0`` by penalizing invalid solutions with ``lam``.

        Parameters
        ----------
        H : dict representing a PUSO.
            The PUSO constraint such that H >= 0. Note that ``H`` will be
            converted to a ``qubovert.PUSO`` object if it is not already,
            thus it must follow the conventions, see ``help(qubovert.PUSO)``.
            Please note that if ``H`` contains any symbols, then ``bounds``
            must be supplied, since they cannot be determined when symbols
            are present.
        lam : float > 0 or sympy.Symbol (optional, defaults to 1).
            Langrange multiplier to penalize violations of the constraint.
        log_trick : bool (optional, defaults to True).
            Whether or not to use the log trick to enforce the inequality
            constraint. See Notes below for more details.
        bounds : two element tuple (optional, defaults to None).
            A tuple ``(min, max)``, the minimum and maximum values that the
            PUSO ``H`` can take. If ``bounds`` is None, then they will be
            calculated (approximately), or if either of the elements of
            ``bounds`` is None, then that element will be calculated
            (approximately).
        suppress_warnings : bool (optional, defaults to False).
            Whether or not to surpress warnings.

        Return
        ------
        self : PCSO.
            Updates the PCSO in place, but returns ``self`` so that operations
            can be strung together.

        Notes
        -----
        - There is no general way to enforce non integer inequality
          constraints. Thus this function is only guarenteed to work for
          integer inequality constraints (ie constraints of the form
          :math:`z_0 + 2z_1 + ... \geq 0`). However, it can be used for non
          integer inequality constraints, but it is recommended that the value
          of ``lam`` be set small, since valid solutions may still recieve a
          penalty to the objective function. For example,

          >>> H = PCSO()
          >>> H.add_constraint_ge_zero(
                  {(0,): -0.5, (): -1.15, (1,): -1.0, (2,): 0.75})
          >>> H
          {(0,): 1.65, (): 4.785, (0, 1): 1.0, (1,): 3.3, (0, 2): -0.75,
           (2,): -2.4749999999999996, ('__a0', 0): 0.5, ('__a0',): 1.65,
           (1, 2): -1.5, ('__a0', 1): 1.0, ('__a0', 2): -0.75}
          >>> H.is_solution_valid(test_sol)
          True
          >>> H.value(test_sol)
          0.01

          {0: -1, 1: -1, 2: 1} is a valid solution to ``H``, but it will still
          cause a nonzero penalty to be added to the objective function.

        - To enforce the inequality constraint, ancilla bits will be
          introduced (labels with `_a`). If ``log_trick`` is ``True``, then
          approximately :math:`\log_2 |\max_z \text{H.value(z)}|`
          ancilla bits will be used. If ``log_trick`` is ``False``, then
          approximately :math:`|\max_z \text{H.value(z)}|` ancilla
          bits will be used.

        """
        H = PUSO(H)
        self._append_constraint("ge", H)
        h = _empty_pcbo(self).add_constraint_ge_zero(
            puso_to_pubo(H), lam=lam, log_trick=log_trick,
            bounds=bounds, suppress_warnings=suppress_warnings
        )
        self._ancilla = h._ancilla
        self += pubo_to_puso(h)
        return self



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_hash.py.

This file contains the hash function that we use to sort keys in the matrices,
used in PUBOMatrix, PUSOMatrix, QUBOMatrix, QUSOMatrix, and all of their
subclasses.

"""

__all__ = 'hash_function',


def hash_function(x):
    """hash_function.

    Function to return (usually) unique hashes for ``x`` such
    that multiple ``x`` s can be ordered. Note that the hash is not consistent
    across Python sessions (except for when ``x`` is an integer)!

    Parameters
    ----------
    x : hashable object.

    Returns
    -------
    res : int.
        If ``x`` is an integer, then ``res == x``. Otherwise,
        ``res == hash(x)``.

    """
    return x if isinstance(x, int) else hash(x)



#   Copyright 2020 Joseph T. Iosue
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""_graph_partitioning.py.

Contains the GraphPartitioning class.
See ``help(qubovert.problems.GraphPartitioning)``.

"""

from qubovert.utils import QUSOMatrix
from qubovert import PCSO
from qubovert.problems import Problem


__all__ = 'GraphPartitioning',


class GraphPartitioning(Problem):
    """GraphPartitioning.

    Class to manage converting (Weighted) Graph Partitioning to and from its
    QUBO and QUSO formluations. Based on the paper "Ising formulations of many
    NP problems", hereforth designated [Lucas].

    The goal of the Graph Partitioning problem is to partition the verticies
    of a graph into two equal subsets such that the number of edges (or the
    total weight of the edges) connecting the two subsets is minimized.

    GraphPartitioning inherits some methods and attributes from the Problem
    class. See ``help(qubovert.problems.Problem)``.

    Example
    -------
    >>> from qubovert.problems import GraphPartitioning
    >>> from any_module import qubo_solver
    >>> # or you can use my bruteforce solver...
    >>> # from qubovert.utils import solve_qubo_bruteforce as qubo_solver
    >>> edges = {("a", "b"), ("a", "c"), ("c", "d"),
                 ("b", "c"), ("e", "f"), ("d", "e")}
    >>> problem = GraphPartitioning(edges)
    >>> Q = problem.to_qubo()
    >>> obj, sol = qubo_solver(Q)
    >>> solution = problem.convert_solution(sol)

    >>> print(solution)
    ({'a', 'b', 'c'}, {'d', 'e', 'f'})

    >>> print(problem.is_solution_valid(solution))
    True

    This is True since the number of vertices in the first partition is equal
    to the number of vertices in the second partition.

    >>> print(obj)
    1

    This is 1 because there is 1 edge connecting the partitions.

    """

    def __init__(self, edges):
        """__init__.

        The goal of the (Weighted) Graph Partitioning problem is to partition
        the vertices of a graph into two equal subsets such that the number of
        edges (or the total weight of the edges) connecting the two subsets is
        minimized. All naming conventions follow the names in the paper
        [Lucas].

        Parameters
        ----------
        edges : set or dict.
            If edges is a set, then it must be a set of two element tuples
            describing the edges of the graph. Ie each tuple is a connection
            between two vertices. If a tuple has a repeated label (for example,
            (2, 2)), it will be ignored.

            If edges is a dict then the keys must be
            two element tuples and the values are the weights associated with
            that edge. If a key has a repeated label (for example, (2, 2)), it
            will be ignored.

        Examples
        --------
        >>> edges = {("a", "b"), ("a", "c")}
        >>> problem = GraphPartitioning(edges)

        >>> edges = {(0, 1), (0, 2)}
        >>> problem = GraphPartitioning(edges)

        >>> edges = {(0, 1): 2, (1, 2): -1}
        >>> problem = GraphPartitioning(edges)

        """
        if isinstance(edges, set):
            self._edges = {k: 1 for k in edges if k[0] != k[1]}
        else:
            self._edges = {k: v for k, v in edges.items() if k[0] != k[1]}

        self._vertices = {y for x in edges for y in x}
        self._vertex_to_index = {x: i for i, x in enumerate(self._vertices)}
        self._index_to_vertex = {i: x for x, i in
                                 self._vertex_to_index.items()}
        self._N = len(self._vertices)

        all_degs = {}
        for e in edges:
            for q in e:
                all_degs[q] = all_degs.setdefault(q, 0) + 1
        self._degree = max(all_degs.values()) if all_degs else 0

    @property
    def E(self):
        """E.

        A copy of the set of edges of the graph. Updating the copy will not
        update the instance set.

        Return
        ------
        E : set of two element tuples.
            A copy of the edge set defining the Graph Partitioning problem.

        """
        return set(self._edges.keys())

    @property
    def V(self):
        """V.

        A copy of the vertex set of the graph. Updating the copy will not
        update the instance set.

        Returns
        -------
        V : set.
            A copy of the set of vertices corresponding to the edge set for the
            Graph Partitioning problem.

        """
        return self._vertices.copy()

    @property
    def weights(self):
        """weights.

        Returns a dictionary mapping the edges of the graph to their associated
        weights.

        Return
        ------
        weights : dict.
            Keys are two element tuples, values are numbers.

        """
        return self._edges.copy()

    @property
    def degree(self):
        """degree.

        The maximum degree of the graph.

        Returns
        -------
        deg : int.
            A copy of the variable of the maximal degree of the graph.

        """
        return self._degree

    @property
    def num_binary_variables(self):
        """num_binary_variables.

        The number of binary variables that the QUBO and QUSO use.

        Return
        ------
        num : integer.
            The number of variables in the QUBO/QUSO formulation.

        """
        return self._N

    def to_quso(self, A=None, B=1):
        r"""to_quso.

        Create and return the graph partitioning problem in QUSO form
        following section 2.2 of [Lucas]. A and B are parameters to enforce
        constraints.

        It is formatted such that the solution to the QUSO formulation is
        equal to the the total number of edges connecting the two
        partitions (or the total weight if we are solving weighted
        partitioning).

        Parameters
        ----------
        A: positive float (optional, defaults to None).
            A enforces the constraints. If it is None, then A will be chosen
            to enforce hard constraints (equation 10 in [Lucas]). Note that
            this may not be optimal for a solver, often hard constraints result
            in unsmooth energy landscapes that are difficult to minimize. Thus
            it may be useful to play around with the magnitude of this value.
        B: positive float (optional, defaults to 1).
            Constant in front of the objective function to minimize. See
            section 2.2 of [Lucas].

        Return
        ------
        L : qubovert.utils.QUSOMatrix object.
            For most practical purposes, you can use QUSOMatrix in the
            same way as an ordinary dictionary. For more information, see
            ``help(qubovert.utils.QUSOMatrix)``.

        Example
        -------
        >>> problem = GraphPartitioning({(0, 1), (1, 2), (0, 3)})
        >>> L = problem.to_quso()

        """
        # all naming conventions follow the paper listed in the docstring
        if A is None:
            A = min(2*self._degree, self._N) * B / 8

        L = QUSOMatrix()

        # encode H_A (equation 8)
        L += PCSO().add_constraint_eq_zero(
            {(i,): 1 for i in range(self._N)}, lam=A)

        # encode H_B (equation 9)
        L += B * sum(self._edges.values()) / 2
        for (u, v), w in self._edges.items():
            L[(self._vertex_to_index[u],
              self._vertex_to_index[v])] -= w * B / 2

        return L

        # slower because we convert to PCSO and then to QUSOMatrix
        # H = PCSO()
        # H.set_mapping(self._vertex_to_index)

        # # encode H_A (equation 8)
        # H.add_constraint_eq_zero({(i,): 1 for i in self._vertices}, lam=A)

        # # encode H_B (equation 9)
        # H += B * sum(self._edges.values()) / 2
        # for e, w in self._edges.items():
        #     H[e] -= w * B / 2

        # return H.to_quso()

    def convert_solution(self, solution, spin=False):
        """convert_solution.

        Convert the solution to the QUBO or QUSO to the solution to the
        Graph Partitioning problem.

        Parameters
        ----------
        solution : iterable or dict.
            The QUBO or QUSO solution output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        res: tuple of sets (partition1, partition2).
            partition1 : set.
                The first partition of verticies.
            partition2 : set.
                The second partition.

        Example
        -------
        >>> edges = {("a", "b"), ("a", "c"), ("c", "d"),
                     ("b", "c"), ("e", "f"), ("d", "e")}
        >>> problem = GraphPartitioning(edges)
        >>> Q = problem.to_qubo()
        >>> obj, sol = solve_qubo(Q)
        >>> print(problem.convert_solution(sol))
        ({'a', 'b', 'c'}, {'d', 'e', 'f'})

        """
        if not isinstance(solution, dict):
            solution = dict(enumerate(solution))

        partition1 = set(
            self._index_to_vertex[i] for i, v in solution.items() if v == 1
        )
        partition2 = set(
            self._index_to_vertex[i] for i, v in solution.items() if v != 1
        )
        return partition1, partition2

    def is_solution_valid(self, solution, spin=False):
        """is_solution_valid.

        Returns whether or not the proposed solution has an equal number of
        vertices in each partition. NOTE: this is impossible if the number of
        edges is odd!

        Parameters
        ----------
        solution : iterable or dict.
            solution can be the output of GraphPartitioning.convert_solution,
            or the  QUBO or QUSO solver output. The QUBO solution output
            is either a list or tuple where indices specify the label of the
            variable and the element specifies whether it's 0 or 1 for QUBO
            (or 1 or -1 for QUSO), or it can be a dictionary that maps the
            label of the variable to is value.
        spin : bool (optional, defaults to False).
            `spin` indicates whether ``solution`` is the solution to the
            boolean {0, 1} formulation of the problem or the spin {1, -1}
            formulation of the problem. This parameter usually does not matter,
            and it will be ignored if possible. The only time it is used is if
            ``solution`` contains all 1's. In this case, it is unclear whether
            ``solution`` came from a spin or boolean formulation of the
            problem, and we will figure it out based on the ``spin`` parameter.

        Return
        ------
        valid : boolean.
            True if the proposed solution is valid, else False.

        """
        not_converted = (
            not isinstance(solution, tuple) or len(solution) != 2 or
            not isinstance(solution[0], set) or
            not isinstance(solution[1], set)
        )

        if not_converted:
            solution = self.convert_solution(solution, spin)

        return len(solution[0]) == len(solution[1])
